#!/usr/bin/env python
"""Functions for tokenizing MEDS datasets.

Here, _tokenization_ refers specifically to the process of converting a longitudinal, irregularly sampled,
continuous time sequence into a temporal sequence at the level that will be consumed by deep-learning models.

All these functions take in _normalized_ data -- meaning data where there are _no longer_ any code modifiers,
as those have been normalized alongside codes into integer indices (in the output code column). The only
columns of concern here thus are `subject_id`, `time`, `code`, `numeric_value`.
"""

from pathlib import Path

import hydra
import polars as pl
from loguru import logger
from MEDS_transforms import PREPROCESS_CONFIG_YAML
from MEDS_transforms.mapreduce.utils import rwlock_wrap, shard_iterator
from MEDS_transforms.utils import hydra_loguru_init, write_lazyframe
from omegaconf import DictConfig, OmegaConf
from safetensors.torch import save_file
from transformers import AutoTokenizer

TOKENIZER = AutoTokenizer.from_pretrained("emilyalsentzer/Bio_ClinicalBERT")

SECONDS_PER_MINUTE = 60.0
SECONDS_PER_HOUR = SECONDS_PER_MINUTE * 60.0
SECONDS_PER_DAY = SECONDS_PER_HOUR * 24.0


def fill_to_nans(col: str | pl.Expr) -> pl.Expr:
    """This function fills infinite and null values with NaN.

    This enables the downstream functions to naturally tensorize data into numpy or Torch tensors.

    Args:
        col: The input column.

    Returns:
        A `pl.Expr` object that fills infinite and null values with NaN.

    Examples:
        >>> print(fill_to_nans("value")) # doctest: +NORMALIZE_WHITESPACE
        .when([(col("value").is_infinite()) |
               (col("value").is_null())]).then(dyn float: NaN).otherwise(col("value"))
        >>> print(fill_to_nans(pl.col("time_delta"))) # doctest: +NORMALIZE_WHITESPACE
        .when([(col("time_delta").is_infinite()) |
               (col("time_delta").is_null())]).then(dyn float: NaN).otherwise(col("time_delta"))
        >>> df = pl.DataFrame({"value": [1.0, float("inf"), None, -float("inf"), 2.0]})
        >>> df.select(fill_to_nans("value").alias("value"))["value"].to_list()
        [1.0, nan, nan, nan, 2.0]
    """

    if isinstance(col, str):
        col = pl.col(col)

    return pl.when(col.is_infinite() | col.is_null()).then(float("nan")).otherwise(col)


def split_static_and_dynamic(df: pl.LazyFrame) -> tuple[pl.LazyFrame, pl.LazyFrame]:
    """This function splits the input data into static and dynamic data.

    Static data is data that has a null time, and dynamic data is everything else.
    For dynamic data, a modality index is added for non-null text values.

    Args:
        df: The input data.

    Returns:
        A tuple of two `pl.LazyFrame` objects, the first being the static data and the second being the
        dynamic data.

    Examples:
        >>> from datetime import datetime
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 2, 2],
        ...     "time": [None, datetime(2021, 1, 1), None, datetime(2021, 1, 2)],
        ...     "code": [100, 101, 200, 201],
        ...     "numeric_value": [1.0, 2.0, 3.0, 4.0],
        ...     "text_value": [None, "fever", None, "cough"]
        ... }).lazy()
        >>> static, dynamic = split_static_and_dynamic(df)
        >>> static.collect()
        shape: (2, 4)
        ┌────────────┬──────┬───────────────┬────────────┐
        │ subject_id ┆ code ┆ numeric_value ┆ text_value │
        │ ---        ┆ ---  ┆ ---           ┆ ---        │
        │ i64        ┆ i64  ┆ f64           ┆ str        │
        ╞════════════╪══════╪═══════════════╪════════════╡
        │ 1          ┆ 100  ┆ 1.0           ┆ null       │
        │ 2          ┆ 200  ┆ 3.0           ┆ null       │
        └────────────┴──────┴───────────────┴────────────┘
        >>> dynamic.collect()
        shape: (2, 6)
        ┌────────────┬─────────────────────┬──────┬───────────────┬────────────┬──────────────┐
        │ subject_id ┆ time                ┆ code ┆ numeric_value ┆ text_value ┆ modality_idx │
        │ ---        ┆ ---                 ┆ ---  ┆ ---           ┆ ---        ┆ ---          │
        │ i64        ┆ datetime[μs]        ┆ i64  ┆ f64           ┆ str        ┆ f32          │
        ╞════════════╪═════════════════════╪══════╪═══════════════╪════════════╪══════════════╡
        │ 1          ┆ 2021-01-01 00:00:00 ┆ 101  ┆ 2.0           ┆ fever      ┆ 1.0          │
        │ 2          ┆ 2021-01-02 00:00:00 ┆ 201  ┆ 4.0           ┆ cough      ┆ 0.0          │
        └────────────┴─────────────────────┴──────┴───────────────┴────────────┴──────────────┘
    """
    static = df.filter(pl.col("time").is_null()).drop("time")
    dynamic = df.filter(pl.col("time").is_not_null())

    # Add modality index for text values
    if "text_value" in df.columns:
        dynamic = dynamic.with_columns(
            [
                pl.when(pl.col("text_value").is_not_null())
                .then(pl.col("text_value").rank("dense") - 1)
                .otherwise(None)
                .cast(pl.Float32)
                .alias("modality_idx")
            ]
        )

    return static, dynamic


def tokenize_text_values(df: pl.DataFrame) -> dict[str, dict]:
    """Tokenize text values and create a mapping of code_modality to tokens.

    Args:
        df: DataFrame containing text values and their corresponding codes and modality indices.

    Returns:
        Dictionary mapping f"{code}_{modality_idx}" to tokenized text.

    Examples:
        >>> df = pl.DataFrame({
        ...     "code": [101, 201],
        ...     "text_value": ["fever", "cough"],
        ...     "modality_idx": [0, 1]
        ... })
        >>> result = tokenize_text_values(df)
        >>> sorted(result.keys())  # Check keys are formatted correctly
        ['0', '1']
        >>> result['0']
        tensor([  101, 10880,   102])
        >>> result['1']
        tensor([  101, 21810,   102])
        >>> # Check empty case
        >>> df_empty = pl.DataFrame({
        ...     "code": [101],
        ...     "text_value": [None],
        ...     "modality_idx": [None]
        ... })
        >>> tokenize_text_values(df_empty)
        {}
    """
    text_mapping = {}

    # Filter to rows with non-null text values
    text_df = df.filter(pl.col("text_value").is_not_null())

    for row in text_df.iter_rows(named=True):
        key = f"{round(row['modality_idx'])}"
        tokens = TOKENIZER(row["text_value"], return_tensors="pt")
        text_mapping[key] = tokens["input_ids"].squeeze()

    return text_mapping


def extract_statics_and_schema(
    df: pl.LazyFrame,
) -> tuple[pl.LazyFrame, dict[str, dict]]:
    """This function extracts static data and schema information (sequence of subject unique times).

    Args:
        df: The input data.

    Returns:
        A tuple containing:
        - A `pl.LazyFrame` object containing the static data and the unique times of the subject
        - A dictionary mapping code_modality to tokenized text

    Examples:
        >>> from datetime import datetime
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 2, 2],
        ...     "time": [None, datetime(2021, 1, 1), datetime(2021, 1, 13),
        ...             None, datetime(2021, 1, 2)],
        ...     "code": [100, 101, 102, 200, 201],
        ...     "numeric_value": [1.0, 2.0, 3.0, 4.0, 5.0],
        ...     "text_value": [None, "fever", "cough", None, "pain"]
        ... }).lazy()
        >>> result_df = extract_statics_and_schema(df)
        >>> result_df.collect()
        shape: (2, 5)
        ┌────────────┬───────────┬───────────────┬─────────────────────┬─────────────────────────────────┐
        │ subject_id ┆ code      ┆ numeric_value ┆ start_time          ┆ time                            │
        │ ---        ┆ ---       ┆ ---           ┆ ---                 ┆ ---                             │
        │ i64        ┆ list[i64] ┆ list[f64]     ┆ datetime[μs]        ┆ list[datetime[μs]]              │
        ╞════════════╪═══════════╪═══════════════╪═════════════════════╪═════════════════════════════════╡
        │ 1          ┆ [100]     ┆ [1.0]         ┆ 2021-01-01 00:00:00 ┆ [2021-01-01 00:00:00, 2021-01-… │
        │ 2          ┆ [200]     ┆ [4.0]         ┆ 2021-01-02 00:00:00 ┆ [2021-01-02 00:00:00]           │
        └────────────┴───────────┴───────────────┴─────────────────────┴─────────────────────────────────┘
    """
    static, dynamic = split_static_and_dynamic(df)

    # This collects static data by subject ID and stores only (as a list) the codes and numeric values
    static_by_subject = static.group_by("subject_id", maintain_order=True).agg("code", "numeric_value")

    # This collects the unique times for each subject
    schema_by_subject = dynamic.group_by("subject_id", maintain_order=True).agg(
        pl.col("time").min().alias("start_time"),
        pl.col("time").unique(maintain_order=True),
    )

    result = static_by_subject.join(schema_by_subject, on="subject_id", how="full", coalesce=True)
    return result


def extract_seq_of_subject_events(
    df: pl.LazyFrame,
) -> tuple[pl.LazyFrame, dict[str, dict]]:
    """This function extracts sequences of subject events, which are sequences of measurements.

    Args:
        df: The input data.

    Returns:
        A tuple containing:
        - A `pl.LazyFrame` object containing the sequences of subject events
        - A dictionary mapping code_modality to tokenized text

    Examples:
        >>> from datetime import datetime
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 2, 2],
        ...     "time": [None, datetime(2021, 1, 1), datetime(2021, 1, 13),
        ...             None, datetime(2021, 1, 2)],
        ...     "code": [100, 101, 102, 200, 201],
        ...     "numeric_value": [1.0, 2.0, 3.0, 4.0, 5.0],
        ...     "text_value": [None, "fever", None, None, "pain"]
        ... }).lazy()
        >>> result_df, text_mapping = extract_seq_of_subject_events(df)
        >>> result_df.collect()
        shape: (2, 5)
        ┌────────────┬─────────────────┬─────────────────┬─────────────────┬─────────────────┐
        │ subject_id ┆ time_delta_days ┆ code            ┆ numeric_value   ┆ modality_idx    │
        │ ---        ┆ ---             ┆ ---             ┆ ---             ┆ ---             │
        │ i64        ┆ list[f32]       ┆ list[list[i64]] ┆ list[list[f64]] ┆ list[list[f32]] │
        ╞════════════╪═════════════════╪═════════════════╪═════════════════╪═════════════════╡
        │ 1          ┆ [NaN, 12.0]     ┆ [[101], [102]]  ┆ [[2.0], [3.0]]  ┆ [[0.0], [NaN]]  │
        │ 2          ┆ [NaN]           ┆ [[201]]         ┆ [[5.0]]         ┆ [[1.0]]         │
        └────────────┴─────────────────┴─────────────────┴─────────────────┴─────────────────┘
        >>> sorted(text_mapping.keys())  # Check text mapping was created
        ['0', '1']
    """
    _, dynamic = split_static_and_dynamic(df)

    # Process text values if they exist
    text_mapping = {}
    if "text_value" in df.columns:
        text_mapping = tokenize_text_values(dynamic.collect())

    time_delta_days_expr = (pl.col("time").diff().dt.total_seconds() / SECONDS_PER_DAY).cast(pl.Float32)

    result = (
        dynamic.group_by("subject_id", "time", maintain_order=True)
        .agg(
            pl.col("code").name.keep(),
            fill_to_nans("numeric_value").name.keep(),
            (fill_to_nans("modality_idx").name.keep() if "text_value" in df.columns else None),
        )
        .group_by("subject_id", maintain_order=True)
        .agg(
            fill_to_nans(time_delta_days_expr).alias("time_delta_days"),
            "code",
            "numeric_value",
            "modality_idx" if "text_value" in df.columns else None,
        )
    )

    return result, text_mapping


@hydra.main(
    version_base=None,
    config_path=str(PREPROCESS_CONFIG_YAML.parent),
    config_name=PREPROCESS_CONFIG_YAML.stem,
)
def main(cfg: DictConfig):
    hydra_loguru_init()
    tokenize(cfg)


def tokenize(cfg: DictConfig):
    """Main function for tokenizing MEDS datasets.

    Examples:
        >>> import tempfile
        >>> import polars as pl
        >>> from datetime import datetime
        >>> from omegaconf import OmegaConf
        >>> from safetensors import safe_open
        >>>
        >>> # Create temporary directory for test data
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     # Create test input data
        ...     test_df = pl.DataFrame({
        ...         "subject_id": [1, 1, 1, 2, 2],
        ...         "time": [None, datetime(2021,1,1), datetime(2021,1,2), None, datetime(2021,1,3)],
        ...         "code": [100, 101, 102, 200, 201],
        ...         "numeric_value": [1.0, 2.0, 3.0, 4.0, 5.0],
        ...         "text_value": [None, "normal", None, None, "abnormal"]
        ...     })
        ...
        ...     # Save test data
        ...     in_fp = Path(tmpdir) / "shard_0.parquet"
        ...     test_df.write_parquet(in_fp)
        ...
        ...     # Create config
        ...     cfg = OmegaConf.create({
        ...         "stage": "tokenize",
        ...         "stage_cfg": {
        ...             "input_dir": str(tmpdir),
        ...             "data_input_dir": str(tmpdir),
        ...             "output_dir": str(tmpdir),
        ...             "file_pattern": "shard_*.parquet",
        ...             "do_sequential": True
        ...         },
        ...         "do_overwrite": True
        ...     })
        ...
        ...     # Run tokenize
        ...     tokenize(cfg)
        ...
        ...     # Verify outputs
        ...     assert (Path(tmpdir) / "schemas" / "shard_0.parquet").exists()
        ...     assert (Path(tmpdir) / "event_seqs" / "shard_0.parquet").exists()
        ...     assert (Path(tmpdir) / "modalities" / "shard_0.safetensors").exists()
        ...
        ...     # Check schema output
        ...     schema_df = pl.read_parquet(Path(tmpdir) / "schemas" / "shard_0.parquet")
        ...     assert len(schema_df) == 2  # Two subjects
        ...     assert all(col in schema_df.columns for col in [
        ...         "subject_id", "code", "numeric_value", "start_time"])
        ...
        ...     # Check event sequences output
        ...     events_df = pl.read_parquet(Path(tmpdir) / "event_seqs" / "shard_0.parquet")
        ...     assert len(events_df) == 2  # Two subjects
        ...     assert all(col in events_df.columns for col in [
        ...         "subject_id", "time_delta_days", "code", "numeric_value", "modality_idx"])
        ...
        ...     # Check event sequences output
        ...     with safe_open(
        ...         Path(tmpdir) / "modalities" / "shard_0.safetensors",
        ...         framework="pt", device="cpu") as f:
        ...         assert set(f.keys()) == {'1', '0'}
        ...         print(f.get_tensor('1'))
        ...         print(f.get_tensor('0'))
        tensor([ 101, 2999,  102])
        tensor([  101, 22832,   102])
    """

    logger.info(
        f"Running with config:\n{OmegaConf.to_yaml(cfg)}\n"
        f"Stage: {cfg.stage}\n\n"
        f"Stage config:\n{OmegaConf.to_yaml(cfg.stage_cfg)}"
    )

    output_dir = Path(cfg.stage_cfg.output_dir)
    if train_only := cfg.stage_cfg.get("train_only", False):
        raise ValueError(f"train_only={train_only} is not supported for this stage.")
    shards_single_output, include_only_train = shard_iterator(cfg)

    for in_fp, out_fp in shards_single_output:
        sharded_path = out_fp.relative_to(output_dir)

        schema_out_fp = output_dir / "schemas" / sharded_path
        event_seq_out_fp = output_dir / "event_seqs" / sharded_path
        text_out_fp = (output_dir / "modalities" / sharded_path).with_suffix(".safetensors")

        logger.info(f"Tokenizing {str(in_fp.resolve())} into schemas at {str(schema_out_fp.resolve())}")

        # Add output path for the LazyFrame to use in compute functions
        rwlock_wrap(
            in_fp,
            schema_out_fp,
            pl.scan_parquet,
            write_lazyframe,
            extract_statics_and_schema,
            do_overwrite=cfg.do_overwrite,
        )

        logger.info(f"Tokenizing {str(in_fp.resolve())} into event_seqs at {str(event_seq_out_fp.resolve())}")

        def write_fn(inputs, out_fp):
            df, text_mapping = inputs
            text_out_fp.parent.mkdir(parents=True, exist_ok=True)
            save_file(text_mapping, text_out_fp)
            write_lazyframe(df, out_fp)

        # Add output path for the LazyFrame to use in compute functions
        rwlock_wrap(
            in_fp,
            event_seq_out_fp,
            pl.scan_parquet,
            write_fn,
            extract_seq_of_subject_events,
            do_overwrite=cfg.do_overwrite,
        )

    logger.info(f"Done with {cfg.stage}")


if __name__ == "__main__":  # pragma: no cover
    main()



from pathlib import Path

import numpy as np
from loguru import logger
from mixins import SeedableMixin
from nested_ragged_tensors.ragged_numpy import JointNestedRaggedTensorDict
from omegaconf import DictConfig

from meds_torch.data.components.pytorch_dataset import PytorchDataset


class RandomWindowPytorchDataset(PytorchDataset):
    """A PyTorch Dataset class that generates random windows on-the-fly for contrastive learning pretraining.

    This class extends PytorchDataset to support random window generation without relying on predefined
    windows.

    Args:
        cfg (DictConfig): Configuration options for the dataset.
        split (str): The data split to use (e.g., 'train', 'validation', 'test').
        min_window_size (int): Minimum size of generated windows.
        max_window_size (int): Maximum size of generated windows.
        n_windows (int): Number of windows to generate for each sample.
    """

    def __init__(self, cfg: DictConfig, split: str):
        super().__init__(cfg, split)
        self.min_window_size = cfg.min_window_size
        self.max_window_size = cfg.max_window_size
        self.n_windows = cfg.n_windows
        self.window_cols = [f"window_{i}" for i in range(cfg.n_windows)]
        self.cfg = cfg
        self.filter_index()

    def filter_index(self):
        filtered_index = []
        for i in range(len(self.index)):
            seq_length = self.index[i][2] - self.index[i][1]
            window_size = self.get_random_window_size(seq_length)
            if window_size < self.min_window_size:
                logger.warning(
                    f"Sequence length {seq_length} is too short to accommodate "
                    f"{self.n_windows} windows of minimum size {self.min_window_size}. "
                    f"Skipping subject {self.index[i][0]}."
                )
            else:
                filtered_index.append(self.index[i])
        if len(filtered_index) < len(self.index):
            logger.warning(f"Filtered data to {len(filtered_index) / len(self.index)}% of original data.")
        self.index = filtered_index

    def get_random_window_size(self, seq_length: int) -> int:
        """Calculates the size of the generated random windows."""
        total_window_size = min(seq_length, self.max_window_size * self.n_windows)
        window_size = total_window_size // self.n_windows
        return window_size

    def generate_random_windows(self, seq_length: int) -> list[tuple[int, int]]:
        """Generate random windows within a sequence.

        This method supports two modes of operation: (a) Random windows: Windows can overlap and be in any
        order. (b) Consecutive random windows: Windows are side by side and in order.

        The window sizes are predetermined to be the same and add up to less than the length of the dataset.

        Args:
            seq_length (int): Length of the sequence to generate windows from.

        Returns:
            List[Tuple[int, int]]: List of (start, end) indices for each window.

        Raises:
            ValueError: If the sequence length is too short to accommodate all windows.
        """
        # Use user-defined window names if provided, otherwise use default names
        window_names = (
            self.cfg.window_names
            if hasattr(self.cfg, "window_names")
            else [f"window_{i}" for i in range(self.n_windows)]
        )

        # Ensure we have the correct number of window names
        if len(window_names) != self.n_windows:
            raise ValueError(
                f"Number of window names ({len(window_names)}) does not match n_windows ({self.n_windows})"
            )

        window_size = self.get_random_window_size(seq_length)
        total_window_size = window_size * self.n_windows
        windows = []

        if self.cfg.get("consecutive_windows", False):
            # Mode (b): Consecutive random windows
            start = np.random.randint(0, seq_length - total_window_size + 1)
            for i in range(self.n_windows):
                end = start + window_size
                windows.append((start, end))
                start = end
        else:
            # Mode (a): Random windows
            for _ in range(self.n_windows):
                start = np.random.randint(0, seq_length - window_size + 1)
                end = start + window_size
                windows.append((start, end))

        # Sort windows by start index to maintain order
        windows.sort(key=lambda x: x[0])

        # Create a dictionary with user-defined or default window names
        named_windows = {name: window for name, window in zip(window_names, windows)}

        return named_windows

    def partition_sequence(self, sequence: dict, windows: list[tuple[int, int]]) -> dict:
        """Partition a sequence into multiple windows.

        Args:
            sequence (dict): The full sequence data.
            windows (List[Tuple[int, int]]): List of (start, end) indices for each window.

        Returns:
            dict: A dictionary with partitioned data for each window.
        """
        partitioned = {}
        for window_name, (start, end) in windows.items():
            partitioned[window_name] = {
                k: v[start:end] if isinstance(v, (list, np.ndarray)) else v for k, v in sequence.items()
            }
        return partitioned

    @SeedableMixin.WithSeed
    def _seeded_getitem(self, idx: int) -> dict:
        """Get a randomly windowed item from the dataset.

        Args:
            idx (int): Index of the item to retrieve.

        Returns:
            dict: A dictionary containing randomly generated windows of the sequence.
        """
        subject_id, _, _ = self.index[idx]
        shard = self.subj_map[subject_id]
        subject_idx = self.subj_indices[subject_id]

        subject_dynamic_data = JointNestedRaggedTensorDict(
            tensors_fp=Path(self.config.data_dir) / "data" / f"{shard}.nrt"
        )[subject_idx]

        full_sequence = self.load_subject(subject_dynamic_data, subject_id, 0, len(subject_dynamic_data))

        seq_len = len(full_sequence["dynamic"])
        windows = self.generate_random_windows(seq_len)
        partitioned_sequence = self.partition_sequence(full_sequence, windows)

        if self.config.do_include_subject_id:
            partitioned_sequence["subject_id"] = subject_id

        return partitioned_sequence

    def collate(self, batch: list[dict]) -> dict:
        """Collate a batch of randomly windowed sequences.

        Args:
            batch (List[dict]): A list of dictionaries, each containing windowed sequences.

        Returns:
            dict: A dictionary with collated data for each window.
        """
        out = {}
        for col in self.window_cols:
            if col in batch[0]:
                out[col] = super().collate([x[col] for x in batch])

        # Collate any additional data that's not part of the windows
        for key in batch[0].keys():
            if key not in self.window_cols and key not in out:
                out[key] = [x[key] for x in batch]

        return out



# TODO: finish the sweep tests
import os
from pathlib import Path

import pytest

from meds_torch.latest_dir import get_latest_directory
from tests.conftest import SUPERVISED_TASK_NAME
from tests.helpers.run_if import RunIf
from tests.helpers.run_sh_command import run_command


def test_train_experiments_fast_dev(tmp_path: Path, meds_dir) -> None:
    """Test running all available experiment configs with `fast_dev_run=True.`

    :param tmp_path: The temporary logging path.
    """
    script = "meds-torch-train"
    args = ["-m"]
    kwargs = {
        "logger": "[]",
        "experiment": "example",
        "++trainer.fast_dev_run": "true",
        "paths.data_dir": str(meds_dir / "triplet_tensors"),
        "paths.meds_cohort_dir": str(meds_dir / "MEDS_cohort"),
        "paths.output_dir": str(tmp_path / "train"),
        "data.task_name": SUPERVISED_TASK_NAME,
        "data.task_root_dir": str(meds_dir / "tasks"),
        "callbacks.model_checkpoint.save_top_k": -1,
    }
    run_command(script=script, args=args, hydra_kwargs=kwargs, test_name="train")
    output_dir = Path(get_latest_directory(tmp_path / "train"))
    assert (output_dir / "hydra_config.yaml").exists()


@RunIf(min_gpus=1)
@pytest.mark.slow
def test_train_experiments(tmp_path: Path, meds_dir) -> None:
    """Test running all available experiment configs with `fast_dev_run=True.`

    :param tmp_path: The temporary logging path.
    """
    script = "meds-torch-train"
    args = ["-m"]
    kwargs = {
        "logger": "[]",
        "experiment": "example",
        "paths.data_dir": str(meds_dir / "triplet_tensors"),
        "paths.meds_cohort_dir": str(meds_dir / "MEDS_cohort"),
        "paths.output_dir": str(tmp_path / "train"),
        "data.task_name": SUPERVISED_TASK_NAME,
        "data.task_root_dir": str(meds_dir / "tasks"),
        "callbacks.model_checkpoint.save_top_k": -1,
        "trainer.max_epochs": 2,
        "trainer.accelerator": "gpu",
    }
    run_command(script=script, args=args, hydra_kwargs=kwargs, test_name="train")
    output_dir = Path(get_latest_directory(tmp_path / "train"))
    assert "last.ckpt" in os.listdir(output_dir / "checkpoints")


@RunIf(min_gpus=1)
@pytest.mark.slow
def test_tune_experiments(tmp_path: Path, meds_dir) -> None:
    """Test running all available experiment configs with `fast_dev_run=True.`

    :param tmp_path: The temporary logging path.
    """
    script = "meds-torch-tune"
    args = ["-m"]
    kwargs = {
        "logger": "[]",
        "hparams_search": "ray_tune",
        "callbacks": "tune_default",
        "trainer": "ray",
        "hparams_search.ray.num_samples": 2,
        "experiment": "example",
        "paths.data_dir": str(meds_dir / "triplet_tensors"),
        "paths.meds_cohort_dir": str(meds_dir / "MEDS_cohort"),
        "paths.output_dir": str(tmp_path / "train"),
        "data.task_name": SUPERVISED_TASK_NAME,
        "data.task_root_dir": str(meds_dir / "tasks"),
        "trainer.max_epochs": 4,
        "trainer.accelerator": "gpu",
    }
    run_command(script=script, args=args, hydra_kwargs=kwargs, test_name="train")
    output_dir = Path(get_latest_directory(tmp_path / "train"))
    subdirectory_count = sum(1 for entry in os.scandir(output_dir / "ray_tune") if entry.is_dir())
    assert subdirectory_count == 2



from typing import Any

from hydra.utils import get_class
from lightning import LightningDataModule
from omegaconf import DictConfig
from torch.utils.data import DataLoader, Dataset

from meds_torch.utils.module_class import Module


def get_dataset(cfg: DictConfig, split):
    dataset_cls = get_class(cfg.dataset_cls)
    return dataset_cls(cfg, split)


class MEDSDataModule(LightningDataModule, Module):
    """`LightningDataModule` for the MEDS pytorch dataset.

    TODO: Add documentation

    A `LightningDataModule` implements 7 key methods:

    ```python     def prepare_data(self):     # Things to do on 1 GPU/TPU (not on every GPU/TPU in DDP).     #
    Pre-process and loading data, save to disk, etc...

    def setup(self, stage): # Things to do on every process in DDP. # Load data, set variables, etc...

    def train_dataloader(self): # return train dataloader

    def val_dataloader(self): # return validation dataloader

    def test_dataloader(self): # return test dataloader

    def predict_dataloader(self): # return predict dataloader

    def teardown(self, stage):     # Called on every process in DDP.     # Clean up after fit or test. ```

    This allows you to share a full dataset without explaining how to download, split, transform and process
    the data.

    Read the docs:
    https://lightning.ai/docs/pytorch/latest/data/datamodule.html
    """

    def __init__(
        self,
        cfg: DictConfig = None,
    ) -> None:
        """Initialize a `MEDSDataModule`."""
        super().__init__()
        self.cfg = cfg

        # this line allows to access init params with 'self.hparams' attribute
        # also ensures init params will be stored in ckpt
        self.save_hyperparameters(logger=False)

        self.data_train: Dataset | None = None
        self.data_val: Dataset | None = None
        self.data_test: Dataset | None = None

    @property
    def num_classes(self) -> int:
        """Get the number of classes.

        :return: The number of MNIST classes (10).
        """
        return 10

    def prepare_data(self) -> None:
        """Download data if needed. Lightning ensures that `self.prepare_data()` is
        called only within a single process on CPU, so you can safely add your
        downloading logic within. In case of multi-node training, the execution of this
        hook depends upon `self.prepare_data_per_node()`.

        Do not use it to assign state (self.x = y).
        """

    def setup(self, stage: str | None = None) -> None:
        """Load data. Set variables: `self.data_train`, `self.data_val`,
        `self.data_test`.

        This method is called by Lightning before `trainer.fit()`, `trainer.validate()`, `trainer.test()`, and
        `trainer.predict()`, so be careful not to execute things like random split twice! Also, it is called
        after `self.prepare_data()` and there is a barrier in between which ensures that all the processes
        proceed to `self.setup()` once the data is prepared and available for use.

        :param stage: The stage to setup. Either `"fit"`, `"validate"`, `"test"`, or `"predict"`. Defaults to
            ``None``.
        """
        # Divide batch size by the number of devices.
        if self.trainer is not None:
            if self.hparams.cfg.dataloader.batch_size % self.trainer.world_size != 0:
                raise RuntimeError(
                    f"Batch size ({self.hparams.cfg.dataloader.batch_size}) is not divisible by "
                    f"the number of devices ({self.trainer.world_size})."
                )
            self.batch_size_per_device = self.hparams.cfg.dataloader.batch_size // self.trainer.world_size

        # load and split datasets only if not loaded already
        if stage == "test":
            self.data_test = get_dataset(self.cfg, split=self.cfg.split_names.test)
        elif stage == "validate":
            self.data_val = get_dataset(self.cfg, split=self.cfg.split_names.validate)
        else:
            self.data_train = get_dataset(self.cfg, split=self.cfg.split_names.train)
            self.data_val = get_dataset(self.cfg, split=self.cfg.split_names.validate)
            self.data_test = get_dataset(self.cfg, split=self.cfg.split_names.test)

    def train_dataloader(self) -> DataLoader[Any]:
        """Create and return the train dataloader.

        :return: The train dataloader.
        """
        return DataLoader(
            dataset=self.data_train,
            shuffle=True,
            collate_fn=self.data_train.collate,
            drop_last=True,
            **self.cfg.dataloader,
        )

    def val_dataloader(self) -> DataLoader[Any]:
        """Create and return the validation dataloader.

        :return: The validation dataloader.
        """
        return DataLoader(
            dataset=self.data_val,
            shuffle=False,
            collate_fn=self.data_val.collate,
            **self.cfg.dataloader,
        )

    def test_dataloader(self) -> DataLoader[Any]:
        """Create and return the test dataloader.

        :return: The test dataloader.
        """
        return DataLoader(
            dataset=self.data_test,
            shuffle=False,
            collate_fn=self.data_test.collate,
            **self.cfg.dataloader,
        )

    def predict_dataloader(self) -> DataLoader[Any]:
        """Create and return the predict dataloader.

        :return: The predict dataloader.
        """
        if self.cfg.predict_dataset == "train":
            return self.train_dataloader()
        elif self.cfg.predict_dataset == "val":
            return self.val_dataloader()
        elif self.cfg.predict_dataset == "test":
            return self.test_dataloader()
        else:
            raise NotImplementedError(
                f"{self.cfg.predict_dataset} not implemented! Use 'train', 'val', or 'test'."
            )

    def state_dict(self) -> dict[Any, Any]:
        """Called when saving a checkpoint. Implement to generate and save the
        datamodule state.

        :return: A dictionary containing the datamodule state that you want to save.
        """
        return {}

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        """Called when loading a checkpoint. Implement to reload datamodule state given
        datamodule `state_dict()`.

        :param state_dict: The datamodule state returned by `self.state_dict()`.
        """


if __name__ == "__main__":
    _ = MEDSDataModule()



from meds_torch.utils.instantiators import (  # noqa
    instantiate_callbacks,
    instantiate_loggers,
)
from meds_torch.utils.logging_utils import log_hyperparameters  # noqa
from meds_torch.utils.pylogger import RankedLogger  # noqa
from meds_torch.utils.rich_utils import enforce_tags, print_config_tree  # noqa
from meds_torch.utils.utils import (  # noqa
    configure_logging,
    get_metric_value,
    task_wrapper,
)



from importlib.resources import files
from typing import Any

import hydra
import torch
from lightning import LightningDataModule, LightningModule, Trainer
from lightning.pytorch.loggers import Logger
from omegaconf import DictConfig

from meds_torch.utils import (
    RankedLogger,
    configure_logging,
    instantiate_loggers,
    log_hyperparameters,
    task_wrapper,
)
from meds_torch.utils.resolvers import setup_resolvers

setup_resolvers()
log = RankedLogger(__name__, rank_zero_only=True)
config_yaml = files("meds_torch").joinpath("configs/eval.yaml")


@task_wrapper
def evaluate(cfg: DictConfig, datamodule=None) -> tuple[dict[str, Any], dict[str, Any]]:
    """Evaluates given checkpoint on a datamodule testset.

    This method is wrapped in optional @task_wrapper decorator, that controls the
    behavior during failure. Useful for multiruns, saving info about the crash, etc.

    Args:
        cfg: DictConfig configuration composed by Hydra.
    Returns:
        Tuple[dict, dict] with metrics and dict with all instantiated objects.
    """
    assert cfg.ckpt_path

    log.info(f"Instantiating datamodule <{cfg.data._target_}>")
    if not datamodule:
        datamodule: LightningDataModule = hydra.utils.instantiate(cfg.data)

    log.info(f"Instantiating model <{cfg.model._target_}>")
    model: LightningModule = hydra.utils.instantiate(cfg.model)
    checkpoint = torch.load(cfg.ckpt_path, map_location="cpu")
    model.load_state_dict(checkpoint["state_dict"])

    log.info("Instantiating loggers...")
    logger: list[Logger] = instantiate_loggers(cfg.get("logger"))

    log.info(f"Instantiating trainer <{cfg.trainer._target_}>")
    trainer: Trainer = hydra.utils.instantiate(cfg.trainer, logger=logger)

    object_dict = {
        "cfg": cfg,
        "datamodule": datamodule,
        "model": model,
        "logger": logger,
        "trainer": trainer,
    }

    if logger:
        log.info("Logging hyperparameters!")
        log_hyperparameters(object_dict)

    log.info("Starting testing!")
    trainer.test(model=model, datamodule=datamodule)

    # for predictions use trainer.predict(...)
    # predictions = trainer.predict(model=model, dataloaders=dataloaders, ckpt_path=cfg.ckpt_path)

    metric_dict = trainer.callback_metrics

    return metric_dict, object_dict


@hydra.main(version_base="1.3", config_path=str(config_yaml.parent.resolve()), config_name=config_yaml.stem)
def main(cfg: DictConfig) -> None:
    """Main entry point for evaluation.

    Args:
        cfg (DictConfig): configuration composed by Hydra.
    """
    # apply extra utilities
    # (e.g. ask for tags if none are provided in cfg, print cfg tree, etc.)
    configure_logging(cfg)

    evaluate(cfg)


if __name__ == "__main__":
    main()



import copy
import glob
import json
import os
import shutil
from collections.abc import Callable
from importlib.resources import files
from pathlib import Path

import hydra
import polars as pl
import ray
from loguru import logger
from omegaconf import DictConfig, OmegaConf, open_dict
from ray import tune
from ray.train import CheckpointConfig, RunConfig, ScalingConfig
from ray.train.lightning import RayLightningEnvironment, prepare_trainer
from ray.train.torch import TorchTrainer

from meds_torch.eval import evaluate
from meds_torch.finetune import initialize_finetune_objects
from meds_torch.train import initialize_train_objects
from meds_torch.utils import RankedLogger, configure_logging, task_wrapper
from meds_torch.utils.resolvers import setup_resolvers

log = RankedLogger(__name__, rank_zero_only=True)

setup_resolvers()
config_yaml = files("meds_torch").joinpath("configs/train.yaml")


@task_wrapper
def ray_tune_runner(cfg: DictConfig, train_func: Callable):
    def objective(config):
        setup_resolvers()
        # Create a new config for this trial
        trial_cfg = copy.deepcopy(cfg)
        # Update the config with the current trial's hyperparameters
        for key, value in config.items():
            OmegaConf.update(trial_cfg, key, value, merge=True)

        # Run the training function
        _ = train_func(trial_cfg)

    # Set up the Ray Tune search space
    search_space = {}
    for key, value in cfg.hparams_search.search_space.items():
        search_space[key] = hydra.utils.instantiate(value)

    scaling_config = ScalingConfig(
        use_gpu=True, resources_per_worker=OmegaConf.to_container(cfg.hparams_search.ray.resources_per_trial)
    )

    run_config = RunConfig(
        checkpoint_config=CheckpointConfig(
            num_to_keep=1,
            checkpoint_score_attribute=cfg.hparams_search.optimized_metric,
            checkpoint_score_order=cfg.hparams_search.direction,
        ),
        name="ray_tune",
        storage_path=cfg.paths.time_output_dir,
    )

    # Define a TorchTrainer without hyper-parameters for Tuner
    ray_trainer = TorchTrainer(
        objective,
        scaling_config=scaling_config,
        run_config=run_config,
    )

    tuner = tune.Tuner(
        ray_trainer,
        param_space={"train_loop_config": search_space},
        tune_config=tune.TuneConfig(
            metric=cfg.hparams_search.optimized_metric,
            mode=cfg.hparams_search.direction,
            num_samples=cfg.hparams_search.ray.num_samples,
            scheduler=hydra.utils.instantiate(cfg.hparams_search.scheduler),
        ),
    )
    analysis = tuner.fit()

    # Return the best trial results
    best_trial = analysis.get_best_result(
        cfg.hparams_search.optimized_metric, cfg.hparams_search.direction, scope="all"
    )
    return analysis, best_trial


def get_checkpoint_path(log_dir, checkpoint_dir_name, time_output_path):
    # Construct the glob pattern
    glob_pattern = os.path.join(
        time_output_path, "ray_tune", f"*{log_dir}*", checkpoint_dir_name, "checkpoint.ckpt"
    )

    # Find the checkpoint file
    checkpoint_files = glob.glob(glob_pattern)

    if len(checkpoint_files) != 1:
        raise ValueError(
            "Expected to find exactly one checkpoint file matching the pattern "
            f"{glob_pattern}, but found {len(checkpoint_files)} files."
        )
    else:
        return checkpoint_files[0]


def train_func(cfg):
    if cfg.hparams_search.train_fn == "train":
        initialize_objects = initialize_train_objects
    elif cfg.hparams_search.train_fn == "finetune":
        initialize_objects = initialize_finetune_objects
    else:
        raise ValueError(f"Invalid train_fn: {cfg.hparams_search.train_fn}, should be 'train' or 'finetune'")
    plugins = [RayLightningEnvironment()]

    object_dict = initialize_objects(cfg, plugins=plugins)
    dm = object_dict["datamodule"]
    model = object_dict["model"]
    trainer = object_dict["trainer"]

    trainer = prepare_trainer(trainer)
    trainer.fit(model, datamodule=dm)


@hydra.main(version_base="1.3", config_path=str(config_yaml.parent.resolve()), config_name=config_yaml.stem)
def main(cfg: DictConfig) -> float | None:
    """Main entry point for training.

    Args:
        cfg: DictConfig configuration composed by Hydra.
    Returns:
        Optional[float] with optimized metric value.
    """
    os.environ["RAY_memory_monitor_refresh_ms"] = cfg.ray_memory_monitor_refresh_ms
    # apply extra utilities
    configure_logging(cfg)

    if cfg.best_config_path:
        if not Path(cfg.best_config_path).exists():
            raise FileNotFoundError(f"Best config file not found at {cfg.best_config_path}")
        logger.info(f"Loading best tuning config from {cfg.best_config_path}")
        with open(Path(cfg.best_config_path)) as config_json:
            best_config = json.load(config_json)["train_loop_config"]
        for key, value in best_config.items():
            OmegaConf.update(cfg, key, value, merge=False)
    else:
        logger.info("No best config provided")

    analysis, best_trial = ray_tune_runner(cfg, train_func=train_func)
    ray.shutdown()

    # return tune results
    results_df = pl.from_dataframe(
        analysis.get_dataframe(
            filter_metric=cfg.hparams_search.optimized_metric, filter_mode=cfg.hparams_search.direction
        )
    )
    results_df.write_parquet(cfg.paths.time_output_dir / "sweep_results.parquet")

    best_model_path = (
        Path(
            best_trial.get_best_checkpoint(
                metric=cfg.hparams_search.optimized_metric, mode=cfg.hparams_search.direction
            ).path
        )
        / "checkpoint.ckpt"
    )

    checkpoint_dir = cfg.paths.time_output_dir / "checkpoints"
    checkpoint_dir.mkdir(parents=True, exist_ok=True)
    shutil.copy(best_model_path, checkpoint_dir / "best_model.ckpt")

    with open(cfg.paths.time_output_dir / "best_config.json", "w") as outfile:
        json.dump(best_trial.config, outfile)

    # Generate summary of results
    summary_results_df_cols = [
        each
        for each in results_df.columns
        if each.startswith("config/")
        or each.startswith("train")
        or each.startswith("val")
        or each.startswith("test")
    ] + ["logdir", "checkpoint_dir_name"]
    summary_df = results_df[summary_results_df_cols]

    # Create a new column 'best_checkpoint_path' using the get_checkpoint_path function
    log_dir_index = summary_df.columns.index("logdir")
    ckpt_dir_name_index = summary_df.columns.index("checkpoint_dir_name")
    checkpoint_paths = summary_df.map_rows(
        lambda x: get_checkpoint_path(x[log_dir_index], x[ckpt_dir_name_index], cfg.paths.time_output_dir)
    )
    summary_df = summary_df.with_columns(best_checkpoint_path=checkpoint_paths.to_series())

    if cfg.get("test"):
        logger.info("Computing Test Results")
        test_results = []
        with open_dict(cfg):
            del cfg.trainer.strategy
            del cfg.callbacks
            cfg.trainer.devices = cfg.test_devices
        datamodule = hydra.utils.instantiate(cfg.data)
        for ckpt_path in summary_df["best_checkpoint_path"].to_list():
            with open_dict(cfg):
                cfg.ckpt_path = ckpt_path
            result, _ = evaluate(cfg, datamodule=datamodule)
            test_results.append(result)
        results = {key: [result[key] for result in test_results] for key in test_results[0].keys()}
        for key, values in results.items():
            summary_df = summary_df.with_columns(pl.Series(values).alias(key))

    logger.info(summary_df)
    summary_df.write_parquet(cfg.paths.time_output_dir / "sweep_results_summary.parquet")

    return best_trial


if __name__ == "__main__":
    main()



import datetime
from dataclasses import asdict, dataclass
from enum import StrEnum
from pathlib import Path

import numpy as np
import polars as pl
import torch
from filelock import FileLock
from loguru import logger
from mixins import SeedableMixin
from nested_ragged_tensors.ragged_numpy import JointNestedRaggedTensorDict
from omegaconf import DictConfig

from meds_torch.data.components.pytorch_dataset import (
    DummyConfig,
    PytorchDataset,
    create_dummy_dataset,
)


def fuse_window_data(windows_data: dict, windows_to_fuse: list[str], fused_window_name: str) -> dict:
    """Fuse multiple windows into a single window, tracking the lengths of original windows.

    Warning:
        - This function assumes that the static data is not prepended to the dynamic data
        - We also assume that static data is the same for all time windows, as it is invariant
            to time windows.

    Args:
        windows_data (dict): Dictionary containing data from multiple windows to be fused
        windows_to_fuse (list[str]): List of window names to fuse in specified order
        fused_window_name (str): Name for the resulting fused window

    Returns:
        dict: Fused window data with length tracking information

    Raises:
        ValueError: If fusion configuration is invalid or data type is unsupported

    Example:
    >>> import torch
    >>> # Create mock data with different types
    >>> windows_data = {
    ...     "pre": {
    ...         # 1D tensor
    ...         "static_values": torch.tensor([1.0, 2.0]),
    ...         # 2D tensor (batch_size=2, seq_len=3, features=2)
    ...         "embeddings": torch.ones(2, 3, 2),
    ...         # List
    ...         "codes": ["A", "B", "C"],
    ...     },
    ...     "post": {
    ...         "static_values": torch.tensor([3.0, 4.0]),
    ...         "embeddings": torch.ones(2, 2, 2) * 2,
    ...         "codes": ["D", "E"],
    ...     }
    ... }
    >>> # Fuse windows
    >>> fused = fuse_window_data(
    ...     windows_data,
    ...     windows_to_fuse=["pre", "post"],
    ...     fused_window_name="fused"
    ... )
    Traceback (most recent call last):
    ...
    ValueError: Unsupported data type <class 'list'> for key codes
    >>> # Remove the list keys
    >>> del windows_data['pre']['codes']
    >>> del windows_data['post']['codes']
    >>> fused = fuse_window_data(
    ...     windows_data,
    ...     windows_to_fuse=["pre", "post"],
    ...     fused_window_name="fused"
    ... )
    >>> # Check lengths tracking
    >>> fused["LENGTHS//static_values"]  # Two windows should select just the first one
    [2, 2]
    >>> fused["LENGTHS//embeddings"]  # Two windows with sequence lengths 3 and 2
    [3, 2]
    >>> # Check only first window's static values (as static data is the same across all windows)
    >>> torch.equal(fused["static_values"], torch.tensor([1.0, 2.0]))
    True
    >>> # Check 2D tensor concatenation along sequence dimension
    >>> torch.equal(fused["embeddings"],
    ...            torch.cat([torch.ones(2, 3, 2),
    ...                      torch.ones(2, 2, 2) * 2], dim=1))
    True
    >>> # Test error handling
    >>> fuse_window_data(windows_data, ["nonexistent"], "fused")
    Traceback (most recent call last):
    ...
    ValueError: Window nonexistent specified in windows_to_fuse not found in data
    >>> # Test error when no fused_window_name
    >>> fuse_window_data(windows_data, ["pre", "post"], None)
    Traceback (most recent call last):
    ...
    ValueError: fused_window_name must not be empty
    """
    if not fused_window_name:
        raise ValueError("fused_window_name must not be empty")

    fused_window = {}
    lengths_tracking = {}

    # Process each window in the specified order
    for window in windows_to_fuse:
        if window not in windows_data:
            raise ValueError(f"Window {window} specified in windows_to_fuse not found in data")

        window_data = windows_data[window]

        # Process each key in the window data
        for key, data in window_data.items():
            # Initialize containers in fused window
            if key not in fused_window:
                lengths_tracking[f"LENGTHS//{key}"] = []

            if not isinstance(data, torch.Tensor):
                raise ValueError(f"Unsupported data type {type(data)} for key {key}")
            # TODO(Oufattole): perform a single concatenation here for efficiency
            logger.warning(f"key: {key} of type: {type(data)} handled as tensor")
            concat_dim = int(len(data.shape) > 1)
            lengths_tracking[f"LENGTHS//{key}"].append(data.shape[concat_dim])
            if key not in fused_window:
                fused_window[key] = data
            else:  # 2D or higher tensor
                if not key.startswith("static_"):
                    fused_window[key] = torch.cat([fused_window[key], data], dim=concat_dim)

    # Convert lists to tensors where appropriate
    for key, value in fused_window.items():
        if isinstance(value, list) and value and isinstance(value[0], torch.Tensor):
            fused_window[key] = torch.cat(value)

    # Add lengths tracking to the fused window
    fused_window.update(lengths_tracking)

    return fused_window


@dataclass
class DummyMultiWindowConfig(DummyConfig):
    """Configuration for MultiWindow dataset"""

    raw_windows_fp: str = None
    cache_dir: str = None
    window_size: int = 30  # Size of each window in days
    window_stride: int = 7  # Stride between windows in days
    min_window_events: int = 1  # Minimum number of events required in a window
    max_windows_per_subject: int | None = None  # Maximum number of windows per subject
    subject_level_sampling: bool = True  # Whether to sample windows at subject level
    default_window_name: str = None
    early_fusion_windows: bool = False  # Whether to fuse windows


def create_dummy_multiwindow_dataset(
    base_dir: str | Path, n_subjects: int = 3, split: str = "train", seed: int | None = 42
) -> DummyMultiWindowConfig:
    """Creates a dummy MultiWindow dataset.

    Args:
        base_dir (str | Path): directory to store the dataset in.
        n_subjects (int, optional): Number of subjects to generate.
        split (str, optional): Is the dataset split. Defaults to "train".
        seed (int | None, optional): Seed used for rng when making the dataset. Defaults to 42.

    Returns:
        DummyMultiWindowConfig: dataset config that can be used to create a MultiWindow dataset.

    Example:
    >>> import tempfile
    >>> _ = pl.Config.set_tbl_width_chars(106)
    >>> with tempfile.TemporaryDirectory() as tmp_dir:
    ...     config = create_dummy_multiwindow_dataset(tmp_dir)
    ...     task_df = pl.read_parquet(Path(config.data_dir) / "task_labels.parquet")
    ...     print(task_df)
    ...     raw_windows_df = pl.read_parquet(Path(config.data_dir) / "raw_windows.parquet")
    ...     print(raw_windows_df.sort("subject_id"))
    shape: (3, 3)
    ┌────────────┬─────────────────────┬───────────────┐
    │ subject_id ┆ prediction_time     ┆ boolean_value │
    │ ---        ┆ ---                 ┆ ---           │
    │ i64        ┆ datetime[μs]        ┆ i64           │
    ╞════════════╪═════════════════════╪═══════════════╡
    │ 0          ┆ 1998-01-01 00:00:00 ┆ 0             │
    │ 1          ┆ 1998-01-01 00:00:00 ┆ 1             │
    │ 2          ┆ 1998-01-01 00:00:00 ┆ 0             │
    └────────────┴─────────────────────┴───────────────┘
    shape: (3, 4)
    ┌────────────┬─────────────────────┬─────────────────────────────────┬─────────────────────────────────┐
    │ subject_id ┆ trigger             ┆ window_1_summary                ┆ window_2_summary                │
    │ ---        ┆ ---                 ┆ ---                             ┆ ---                             │
    │ i64        ┆ datetime[μs]        ┆ struct[4]                       ┆ struct[4]                       │
    ╞════════════╪═════════════════════╪═════════════════════════════════╪═════════════════════════════════╡
    │ 0          ┆ 1998-01-01 00:00:00 ┆ {0,1998-01-01 00:00:00,1995-01… ┆ {0,1998-01-01 00:00:00,1998-01… │
    │ 1          ┆ 1998-01-01 00:00:00 ┆ {1,1998-01-01 00:00:00,1995-01… ┆ {1,1998-01-01 00:00:00,1998-01… │
    │ 2          ┆ 1998-01-01 00:00:00 ┆ {2,1998-01-01 00:00:00,1995-01… ┆ {2,1998-01-01 00:00:00,1998-01… │
    └────────────┴─────────────────────┴─────────────────────────────────┴─────────────────────────────────┘
    """
    config = create_dummy_dataset(base_dir, n_subjects, split, seed)
    raw_windows_fp = Path(base_dir) / "raw_windows.parquet"
    cache_dir = Path(base_dir) / "cache"
    task_df = pl.read_parquet(Path(config.data_dir) / "task_labels.parquet")
    static_df = pl.read_parquet(Path(config.data_dir) / "schema/train/shard_0.parquet")
    window_1 = task_df.hstack(static_df.drop("subject_id")).select(
        "subject_id",
        "prediction_time",
        pl.col("start_time").alias("timestamp_at_start"),
        pl.col("prediction_time").alias("timestamp_at_end"),
    )
    window_2 = task_df.hstack(static_df.drop("subject_id")).select(
        "subject_id",
        "prediction_time",
        pl.col("prediction_time").alias("timestamp_at_start"),
        pl.col("time").list.max().alias("timestamp_at_end"),
    )
    raw_windows_df = task_df.select("subject_id", pl.col("prediction_time").alias("trigger")).with_columns(
        window_1.to_struct().alias("window_1_summary"), window_2.to_struct().alias("window_2_summary")
    )
    raw_windows_df.write_parquet(raw_windows_fp)

    return DummyMultiWindowConfig(raw_windows_fp=raw_windows_fp, cache_dir=cache_dir, **asdict(config))


def get_window_indexes(static_df: pl.DataFrame, windows_df: pl.DataFrame) -> pl.DataFrame:
    """Computes the start and end indexes of time windows for each entry in the provided DataFrame. This
    function assumes that the "time" in `timestamps_series` is sorted. It finds the index of timestamps that
    fall between 'start' and 'end' times specified in `windows_df`.

    Parameters:
    - static_df (pl.Series): A Polars dataframe containing sorted datetime values for each subject.
    - windows_df (pl.DataFrame): A DataFrame with columns 'name', 'start', and 'end' specifying the time
        windows.

    Returns:
    - pl.DataFrame: A DataFrame with the original columns of `windows_df` plus for each <COL_NAME>.start
        and <COL_NAME>.end in windows_df, the column '<COL_NAME>.start_idx' and '<COL_NAME>.end_idx'
        indicating the inclusive index range of timestamps within each window is added.

    Example:
    >>> import pprint
    >>> _ = pl.Config.set_tbl_width_chars(106)
    >>> timeseries_df = pl.DataFrame({
    ...     "subject_id": [1, 2],
    ...     "time": [
    ...         pl.Series(["1978-03-09 00:00:00", "2010-05-26 02:30:56", "2010-05-26 04:51:52"]
    ...             ).str.strptime(pl.Datetime),
    ...         pl.Series(["1970-03-09 00:00:00", "1972-05-26 02:30:56", "1975-05-26 04:51:52"]
    ...             ).str.strptime(pl.Datetime)
    ...     ]
    ... })
    >>> windows_df = pl.DataFrame({
    ...     "subject_id": [1, 2],
    ...     "pre.start": [["1978-03-09 00:00:00", "2010-05-26 02:30:56"], ["1969-05-26 02:30:56"]],
    ...     "pre.end": [["2010-05-26 02:30:56", "2010-05-26 04:51:52"], ["1971-05-26 04:51:52"]],
    ...     "post.start": [["1978-03-09 00:00:00", "2010-05-26 02:30:56"], ["1971-05-26 02:30:56"]],
    ...     "post.end": [["2010-05-26 02:30:56", "2010-05-26 04:51:52"], ["1980-05-26 04:51:52"]],
    ... }).with_columns([
    ...     pl.col("pre.start").list.eval(pl.element().str.to_datetime()),
    ...     pl.col("pre.end").list.eval(pl.element().str.to_datetime()),
    ...     pl.col("post.start").list.eval(pl.element().str.to_datetime()),
    ...     pl.col("post.end").list.eval(pl.element().str.to_datetime()),
    ... ])
    >>> timeseries_df
    shape: (2, 2)
    ┌────────────┬─────────────────────────────────┐
    │ subject_id ┆ time                            │
    │ ---        ┆ ---                             │
    │ i64        ┆ list[datetime[μs]]              │
    ╞════════════╪═════════════════════════════════╡
    │ 1          ┆ [1978-03-09 00:00:00, 2010-05-… │
    │ 2          ┆ [1970-03-09 00:00:00, 1972-05-… │
    └────────────┴─────────────────────────────────┘
    >>> windows_df.shape
    (2, 5)
    >>> get_window_indexes(
    ...     timeseries_df, windows_df).select("subject_id", pl.col("^.*_idx$")).sort("subject_id")
    shape: (2, 5)
    ┌────────────┬───────────────┬─────────────┬────────────────┬──────────────┐
    │ subject_id ┆ pre.start_idx ┆ pre.end_idx ┆ post.start_idx ┆ post.end_idx │
    │ ---        ┆ ---           ┆ ---         ┆ ---            ┆ ---          │
    │ i64        ┆ list[u32]     ┆ list[u32]   ┆ list[u32]      ┆ list[u32]    │
    ╞════════════╪═══════════════╪═════════════╪════════════════╪══════════════╡
    │ 1          ┆ [0, 1]        ┆ [1, 2]      ┆ [0, 1]         ┆ [1, 2]       │
    │ 2          ┆ [0]           ┆ [1]         ┆ [1]            ┆ [3]          │
    └────────────┴───────────────┴─────────────┴────────────────┴──────────────┘
    >>> single_event_per_subject_window_df = (windows_df
    ...                                       .explode(pl.exclude("subject_id"))
    ...                                       .group_by("subject_id").first()
    ...                                       .group_by("subject_id").agg(pl.all()))
    >>> pprint.pprint({k: v.to_list() for k,v in single_event_per_subject_window_df
    ...                  .sort("subject_id").to_dict().items()})
    {'post.end': [[datetime.datetime(2010, 5, 26, 2, 30, 56)],
                  [datetime.datetime(1980, 5, 26, 4, 51, 52)]],
     'post.start': [[datetime.datetime(1978, 3, 9, 0, 0)],
                    [datetime.datetime(1971, 5, 26, 2, 30, 56)]],
     'pre.end': [[datetime.datetime(2010, 5, 26, 2, 30, 56)],
                 [datetime.datetime(1971, 5, 26, 4, 51, 52)]],
     'pre.start': [[datetime.datetime(1978, 3, 9, 0, 0)],
                   [datetime.datetime(1969, 5, 26, 2, 30, 56)]],
     'subject_id': [1, 2]}
    >>> get_window_indexes(
    ...     timeseries_df, single_event_per_subject_window_df
    ... ).select("subject_id", pl.col("^.*_idx$")).sort("subject_id")
    shape: (2, 5)
    ┌────────────┬───────────────┬─────────────┬────────────────┬──────────────┐
    │ subject_id ┆ pre.start_idx ┆ pre.end_idx ┆ post.start_idx ┆ post.end_idx │
    │ ---        ┆ ---           ┆ ---         ┆ ---            ┆ ---          │
    │ i64        ┆ list[u32]     ┆ list[u32]   ┆ list[u32]      ┆ list[u32]    │
    ╞════════════╪═══════════════╪═════════════╪════════════════╪══════════════╡
    │ 1          ┆ [0]           ┆ [1]         ┆ [0]            ┆ [1]          │
    │ 2          ┆ [0]           ┆ [1]         ┆ [1]            ┆ [3]          │
    └────────────┴───────────────┴─────────────┴────────────────┴──────────────┘
    """
    datetime_cols = [col for col in windows_df.columns if col.endswith(".start") or col.endswith(".end")]
    output_cols = [f"{col}_idx" for col in datetime_cols]
    windows_df = windows_df.join(how="inner", other=static_df, on="subject_id")
    expr = [
        (pl.col("time").explode().search_sorted(pl.col(col).explode()).alias(f"{col}_idx"))
        for col in datetime_cols
    ]
    windows_df = windows_df.group_by(pl.col("subject_id")).agg(expr)
    if output_cols and not isinstance(windows_df.schema[output_cols[0]], pl.List):
        windows_df = windows_df.group_by("subject_id").agg(pl.all())
    return windows_df


def cache_window_indexes(cfg: DictConfig, split: str, static_dfs) -> pl.DataFrame:
    """Caches window indexes for the given split of the dataset.

    Args:
        cfg (DictConfig): _description_
        split (str): _description_
        static_dfs (_type_): _description_

    Returns:
        pl.DataFrame: _description_

    Example:
    >>> import tempfile
    >>> import pprint
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     config = create_dummy_multiwindow_dataset(tmpdir)
    ...     static_df = pl.read_parquet(Path(config.data_dir) / "schema/train/shard_0.parquet")
    ...     raw_windows_df = pl.read_parquet(Path(config.data_dir) / "raw_windows.parquet")
    ...     cache_window_indexes(config, "train", {"shard_0": static_df})
    ...     cached_window_df = pl.read_parquet(Path(config.cache_dir) / "train.parquet").sort("subject_id")
    >>> pprint.pprint({k: v.to_list() for k,v in cached_window_df.to_dict().items()})
    {'subject_id': [0, 1, 2],
     'window_1_summary.end_idx': [[3], [3], [3]],
     'window_1_summary.start_idx': [[0], [0], [0]],
     'window_2_summary.end_idx': [[4], [4], [4]],
     'window_2_summary.start_idx': [[3], [3], [3]]}
    """
    window_df = pl.read_parquet(cfg.raw_windows_fp)
    window_cols = [col for col in window_df.columns if col.endswith("_summary")]
    col = "pre.start_summary"
    exprs = [pl.col("subject_id")]
    for col in window_cols:
        for side in ("start", "end"):
            parsed_col = col.split(".")[0] + f".{side}"
            exprs.append(pl.col(col).struct.field(f"timestamp_at_{side}").alias(parsed_col))
    window_df = window_df.select(exprs)
    window_df = window_df.group_by("subject_id").agg(pl.all())
    static_df = pl.concat(static_dfs.values()).select("subject_id", "time")
    cached_window_df = get_window_indexes(static_df, window_df)
    cache_window_fp = Path(cfg.cache_dir) / f"{split}.parquet"
    cache_window_fp.parent.mkdir(parents=True, exist_ok=True)
    cached_window_df.write_parquet(cache_window_fp)


class MultiWindowSamplingStrategy(StrEnum):
    """Enumeration of sampling strategies for multi-window datasets.

    Attributes:
        RANDOM: Randomly sample and window from the dataset, partitioning it into separate
            subwindows.
        PREDEFINED: Use predefined windows around events to sample from the dataset.
    """

    RANDOM = "random"
    PREDEFINED = "predefined"


class MultiWindowPytorchDataset(SeedableMixin, torch.utils.data.Dataset):
    """A Multi-Window PyTorch Dataset class for contrastive learning pretraining.

    This class extends the functionality of a standard PyTorch Dataset to support multiple time windows for
    each subject. It's designed to work with medical event data, where each subject may have multiple relevant
    time windows for analysis.

    The dataset can be configured to sample at the subject level or at the window level, allowing for flexible
    data loading strategies in contrastive learning scenarios.

    Args:
        cfg (DictConfig): Configuration options for the dataset.
        split (str): The data split to use (e.g., 'train', 'validation', 'test').

    Attributes:
        config (DictConfig): The configuration object for the dataset.
        split (str): The current data split being used.
        pytorch_dataset (PytorchDataset): The underlying PyTorch dataset.
        window_cols (list): List of window column names.
        index (list): List of dictionaries, each representing a subject or a window,
            depending on the sampling strategy.

    Example:
    >>> import tempfile
    >>> from pathlib import Path
    >>> import torch
    >>> import polars as pl
    >>> from omegaconf import OmegaConf
    >>>
    >>> # Create dummy dataset in a temporary directory
    >>> with tempfile.TemporaryDirectory() as tmp_dir:
    ...     # Generate dummy config with sample data
    ...     config = create_dummy_multiwindow_dataset(tmp_dir)
    ...     cfg = config
    ...
    ...     # Initialize the dataset
    ...     dataset = MultiWindowPytorchDataset(cfg, split="train")
    ...
    ...     # Check the available windows
    ...     print(f"Window columns: {dataset.window_cols}")
    ...
    ...     # Get a sample item
    ...     sample = dataset[0]
    ...
    ...     # Examine the structure of the first window
    ...     window_name = dataset.window_cols[0]
    ...     print(f"\\nStructure of {window_name} window:")
    ...     for key, value in sample[window_name].items():
    ...         if isinstance(value, dict):
    ...             print(f"{key}:")
    ...             for subkey, subvalue in value.items():
    ...                 print(f"  {subkey}: {type(subvalue)}")
    ...         else:
    ...             print(f"{key}: {type(value)}")
    ...
    ...     # Create a dataloader and get a batch
    ...     dataloader = torch.utils.data.DataLoader(
    ...         dataset,
    ...         batch_size=2,
    ...         collate_fn=dataset.collate
    ...     )
    ...     batch = next(iter(dataloader))
    Window columns: ['window_1_summary', 'window_2_summary']
    <BLANKLINE>
    Structure of window_1_summary window:
    static_indices: <class 'list'>
    static_values: <class 'list'>
    start_idx: <class 'int'>
    end_idx: <class 'int'>
    dynamic: <class 'nested_ragged_tensors.ragged_numpy.JointNestedRaggedTensorDict'>
    start_time: <class 'datetime.datetime'>
    end_time: <class 'datetime.datetime'>
    """

    def __init__(self, cfg: DictConfig, split: str):
        """Initialize the MultiWindowPytorchDataset.

        This method sets up the dataset by loading or creating cached window indexes, initializing the
        underlying PytorchDataset, and preparing the index based on the specified sampling strategy.

        Args:
            cfg (DictConfig): Configuration options for the dataset.
            split (str): The data split to use (e.g., 'train', 'validation', 'test').
        """
        super().__init__()

        self.config = cfg
        self.split = split
        self.pytorch_dataset = PytorchDataset(cfg, split)
        cached_windows_fp = Path(cfg.cache_dir) / f"{split}.parquet"
        with FileLock(f"{cached_windows_fp}.lock"):
            if cached_windows_fp.exists():
                window_df = pl.read_parquet(cached_windows_fp)
            else:
                logger.info("No cached windows found. Caching now.")
                cache_window_indexes(cfg, split, self.pytorch_dataset.static_dfs)
                window_df = pl.read_parquet(cached_windows_fp)
        self.window_cols = sorted(
            list({col.split(".")[0] for col in window_df.columns if col.endswith("_idx")})
        )
        window_df = self.filter_invalid_window(window_df)
        if self.config.subject_level_sampling:
            # index by subject_id
            self.index = window_df.to_dicts()
        else:
            # index by windows
            self.index = window_df.explode(
                [col for col in window_df.columns if col.endswith("_idx")]
            ).to_dicts()

    def filter_invalid_window(cls, window_df: pl.DataFrame) -> pl.DataFrame:
        """Filter out invalid windows if st index >= end index."""
        window_df = window_df.explode(pl.exclude("subject_id"))
        num_rows = window_df.shape[0]
        window_cols = {col[:-10] for col in window_df.columns if col.endswith("start_idx")}
        window_exprs = [
            pl.col(f"{window_col}.end_idx") > pl.col(f"{window_col}.start_idx") for window_col in window_cols
        ]
        window_df = window_df.filter(window_exprs)
        if window_df.shape[0] != num_rows:
            logger.warning(f"Filtered out {num_rows - window_df.shape[0]} invalid windows.")
        window_df = window_df.group_by("subject_id").agg(pl.all())
        return window_df

    @property
    def subject_ids(self) -> list[int]:
        return [x[0] for x in self.index]

    def __len__(self):
        return len(self.index)

    @property
    def has_task(self) -> bool:
        return self.config.task_name is not None

    @property
    def max_seq_len(self) -> int:
        return self.config.max_seq_len

    def collate(self, batch: dict[str : np.array]) -> dict[str : torch.Tensor]:
        """Collate a batch of data samples into a single batch.

        This method is responsible for combining multiple data samples into a single batch that can be
        processed by a PyTorch model. It handles both window-specific data and any additional data that might
        be present.

        Args:
            batch (dict[str, np.array]): A dictionary of arrays, each representing a batch of data for a
                specific feature.

        Returns:
            dict[str, torch.Tensor]: A dictionary of tensors, representing the collated batch data.
        """
        out = {}
        for col in self.window_cols:
            out[col] = self.pytorch_dataset.collate([x[col] for x in batch])
        for key in batch[0].keys():
            if key not in self.window_cols:
                if isinstance(batch[0][key], datetime.datetime):
                    out[key] = [item[key] for item in batch]
                else:
                    out[key] = torch.Tensor([item[key] for item in batch])
        # Fuse windows
        if self.config.early_fusion_windows:
            if self.config.do_prepend_static_data:
                raise ValueError("early fusion of windows with do_prepend_static_data is not supported.")
            out[self.config.early_fusion_window_name] = fuse_window_data(
                out, self.config.early_fusion_windows, self.config.early_fusion_window_name
            )
        if self.config.default_window_name:
            for key in out[self.config.default_window_name].keys():
                out[key] = out[self.config.default_window_name][key]

        return out

    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        """Get a single item from the dataset.

        This method retrieves data for a single subject, returning a dictionary where each key represents a
        window, and the corresponding value is the output of PytorchDataset.__getitem__ for that window.

        Args:     idx (int): The index of the item to retrieve.

        Returns:
            dict[str, dict]: A dictionary where keys are window names (e.g., 'pre', 'post') and
                values are dictionaries containing the data for each window.
                The structure of each window's data typically includes:
                    - 'static_indices': List of static categorical metadata elements.
                    - 'static_values': List of static numerical metadata elements.
                    - 'dynamic':
                        Dictionary containing:
                        - 'time_delta_days': List of time deltas between events.
                        - 'dim1/code': List of dynamic categorical metadata elements.
                        - 'dim1/numeric_value': List of dynamic numerical metadata elements.
        """
        return self._seeded_getitem(idx)

    @SeedableMixin.WithSeed
    def _seeded_getitem(self, idx: int) -> dict[str, list[float]]:
        """Seedable version of __getitem__.

        This method is a seedable version of __getitem__, allowing for reproducible data retrieval. It handles
        both subject-level and window-level sampling, returning a dictionary of window data for the selected
        subject.

        Args:
            idx (int): The index of the item to retrieve.

        Returns:
            dict[str, dict]: A dictionary where keys are window names (e.g., 'pre', 'post') and
                values are dictionaries containing the data for each window,
                as returned by PytorchDataset.load_subject().
        """
        if self.config.subject_level_sampling:
            # index by subject_id
            subject_data = self.index[idx]
            num_windows = len(subject_data[f"{self.window_cols[0]}.start_idx"])
            selected_window_idx = np.random.choice(num_windows)
            windows = {
                col: [
                    subject_data[f"{col}.start_idx"][selected_window_idx],
                    subject_data[f"{col}.end_idx"][selected_window_idx],
                ]
                for col in self.window_cols
            }
        else:
            # index by windows
            subject_data = self.index[idx]
            windows = {
                col: [subject_data[f"{col}.start_idx"], subject_data[f"{col}.end_idx"]]
                for col in self.window_cols
            }
        subject_id = subject_data["subject_id"]

        shard = self.pytorch_dataset.subj_map[subject_id]
        subject_idx = self.pytorch_dataset.subj_indices[subject_id]
        out = {}
        for window in self.window_cols:
            st, end = windows[window][0], windows[window][1]
            subject_dynamic_data = JointNestedRaggedTensorDict(
                tensors_fp=Path(self.config.data_dir) / "data" / f"{shard}.nrt"
            )[subject_idx, st:end]

            if st >= end:
                raise ValueError(f"start index {st} >= end index {end}")

            out[window] = self.pytorch_dataset.load_subject(subject_dynamic_data, subject_id, st, end)

        if self.config.do_include_subject_id:
            out["subject_id"] = subject_id
        return out



import torch
import torchmetrics
from loguru import logger
from omegaconf import DictConfig
from torch import nn

from meds_torch.models import BACKBONE_EMBEDDINGS_KEY, MODEL_BATCH_LOSS_KEY
from meds_torch.models.base_model import BaseModule


class ValueForecastingModule(BaseModule):
    def __init__(self, cfg: DictConfig):
        super().__init__(cfg)
        # pretraining model components
        self.presence_projection = nn.Linear(cfg.token_dim, cfg.vocab_size)
        self.presence_criterion = nn.MSELoss()

        # logging components
        self.train_presence_mse = torchmetrics.MeanSquaredError()
        self.train_value_mse = torchmetrics.MeanSquaredError()

        self.val_presence_mse = torchmetrics.MeanSquaredError()
        self.val_value_mse = torchmetrics.MeanSquaredError()

        self.test_presence_mse = torchmetrics.MeanSquaredError()
        self.test_value_mse = torchmetrics.MeanSquaredError()

        self.value_projection = nn.Linear(cfg.token_dim, cfg.vocab_size)
        self.value_criterion = nn.MSELoss()

    @staticmethod
    def set_target(empty_target, row_indices, col_indices, value=1):
        try:
            empty_target[row_indices, col_indices] = 1
        except IndexError:
            logger.warning("Index out of bounds, doing inefficient loop")
            for i, j in zip(row_indices, col_indices):
                empty_target[i, j] = 1

    def forward(self, batch):
        forecast_window_data = batch[self.cfg.forecast_window_name]
        batch[self.cfg.input_window_name] = self.model(self.input_encoder(batch[self.cfg.input_window_name]))

        numeric_values = forecast_window_data["numeric_value"]
        codes = forecast_window_data["code"]
        vocab_size = self.cfg.vocab_size

        with torch.no_grad():
            # create presence target and value target
            presence_target = torch.zeros(
                (codes.shape[0], vocab_size), dtype=torch.float32, device=codes.device
            )
            presence_target = presence_target.scatter_(
                dim=1,
                index=codes.to(torch.int64),
                src=torch.ones_like(codes, dtype=torch.float32),
                reduce="add",
            ).clamp_(max=1)

            # create value target
            numeric_value_mask = forecast_window_data["numeric_value_mask"]
            numeric_value_codes = codes * numeric_value_mask
            value_target = torch.zeros(
                (numeric_value_codes.shape[0], vocab_size), dtype=torch.float32, device=codes.device
            )
            value_target = value_target.scatter_(
                dim=1, index=codes.to(torch.int64), src=numeric_values, reduce="add"
            )
            count_target = torch.zeros_like(value_target)
            count_target = count_target.scatter_(
                dim=1, index=codes.to(torch.int64), src=torch.ones_like(numeric_values), reduce="add"
            )
            value_target = torch.where(count_target > 0, value_target / count_target, value_target)

            # Set the 0 index to zero to ignore mask tokens
            value_target[:, 0] = 0

        value_forecast = self.value_projection(batch[self.cfg.input_window_name][BACKBONE_EMBEDDINGS_KEY])
        presence_forecast = self.presence_projection(
            batch[self.cfg.input_window_name][BACKBONE_EMBEDDINGS_KEY]
        )

        value_loss = self.value_criterion(value_forecast, value_target)
        presence_loss = self.presence_criterion(presence_forecast, presence_target)
        loss = value_loss + presence_loss

        batch["MODEL//VALUE_TARGET"] = value_target
        batch["MODEL//PRESENCE_TARGET"] = presence_target
        batch["MODEL//VALUE_FORECAST"] = value_forecast
        batch["MODEL//PRESENCE_FORECAST"] = presence_forecast
        batch["MODEL//VALUE_LOSS"] = value_loss
        batch["MODEL//PRESENCE_LOSS"] = presence_loss
        batch[MODEL_BATCH_LOSS_KEY] = loss

        return batch

    def training_step(self, batch):
        output = self.forward(batch)

        self.train_presence_mse(output["MODEL//PRESENCE_FORECAST"], output["MODEL//PRESENCE_TARGET"])
        self.train_value_mse(output["MODEL//VALUE_FORECAST"], output["MODEL//VALUE_TARGET"])

        self.log("train/loss", output[MODEL_BATCH_LOSS_KEY], batch_size=self.cfg.batch_size)
        return output[MODEL_BATCH_LOSS_KEY]

    def validation_step(self, batch):
        output = self.forward(batch)

        self.val_presence_mse(output["MODEL//PRESENCE_FORECAST"], output["MODEL//PRESENCE_TARGET"])
        self.val_value_mse(output["MODEL//VALUE_FORECAST"], output["MODEL//VALUE_TARGET"])

        self.log("val/loss", output[MODEL_BATCH_LOSS_KEY], batch_size=self.cfg.batch_size)
        return output[MODEL_BATCH_LOSS_KEY]

    def test_step(self, batch):
        output = self.forward(batch)

        self.test_presence_mse(output["MODEL//PRESENCE_FORECAST"], output["MODEL//PRESENCE_TARGET"])
        self.test_value_mse(output["MODEL//VALUE_FORECAST"], output["MODEL//VALUE_TARGET"])

        self.log("test/loss", output[MODEL_BATCH_LOSS_KEY], batch_size=self.cfg.batch_size)

        return output[MODEL_BATCH_LOSS_KEY]

    def on_train_epoch_end(self):
        self.log(
            "train_presence_mse",
            self.train_presence_mse,
            on_epoch=True,
        )
        self.log(
            "train_value_mse",
            self.train_value_mse,
            on_epoch=True,
        )

    def on_validation_epoch_end(self):
        self.log(
            "val_presence_mse",
            self.val_presence_mse,
            on_epoch=True,
        )
        self.log(
            "val_value_mse",
            self.val_value_mse,
            on_epoch=True,
        )

    def on_test_epoch_end(self):
        self.log(
            "test_presence_mse",
            self.test_presence_mse,
            on_epoch=True,
        )
        self.log(
            "test_value_mse",
            self.test_value_mse,
            on_epoch=True,
        )



import rootutils

root = rootutils.setup_root(__file__, dotenv=True, pythonpath=True, cwd=True)

import functools
import operator
from pathlib import Path

import polars as pl
import pytest
import torch
from omegaconf import open_dict
from torch.utils.data import DataLoader

from meds_torch.data.components.multiwindow_pytorch_dataset import (
    MultiWindowPytorchDataset,
)
from meds_torch.data.components.pytorch_dataset import PytorchDataset
from meds_torch.data.components.random_windows_pytorch_dataset import (
    RandomWindowPytorchDataset,
)
from meds_torch.data.datamodule import MEDSDataModule
from tests.conftest import SUPERVISED_TASK_NAME, create_cfg


@pytest.mark.parametrize(
    "collate_type",
    [
        "triplet",
        "triplet_prompt",
        "eic",
    ],
)
def test_pytorch_dataset(meds_dir, collate_type):
    cfg = create_cfg(overrides=[], meds_dir=meds_dir)
    cfg.data.collate_type = collate_type
    cfg.data.tokenizer = "emilyalsentzer/Bio_ClinicalBERT"
    pyd = PytorchDataset(cfg.data, split="train")
    assert not pyd.has_task
    item = pyd[0]
    assert item.keys() == {"static_indices", "static_values", "dynamic"}
    batch = pyd.collate([pyd[i] for i in range(2)])
    if collate_type == "triplet":
        assert batch.keys() == {
            "mask",
            "static_mask",
            "code",
            "numeric_value",
            "time_delta_days",
            "numeric_value_mask",
        }
    elif collate_type == "triplet_prompt":
        assert batch.keys() == {
            "mask",
            "static_mask",
            "code",
            "numeric_value",
            "time_delta_days",
            "numeric_value_mask",
        }
    elif collate_type == "eic":
        assert batch.keys() == {
            "mask",
            "static_mask",
            "code",
            "numeric_value",
            "time_delta_days",
            "numeric_value_mask",
        }
    else:
        raise NotImplementedError(f"{collate_type} not implemented")


@pytest.mark.parametrize("collate_type", ["triplet", "triplet_prompt", "eic"])
def test_pytorch_dataset_with_supervised_task(meds_dir, collate_type):
    cfg = create_cfg(overrides=[], meds_dir=meds_dir, supervised=True)
    with open_dict(cfg):
        cfg.data.collate_type = collate_type
        cfg.data.dataloader.batch_size = 70
    assert Path(cfg.data.task_label_path).exists(), f"Path does not exist: {cfg.data.task_label_path}"

    pyd = PytorchDataset(cfg.data, split="train")
    assert len(pyd) == 70
    assert pyd.has_task
    item = pyd[0]
    assert item.keys() == {"static_indices", "static_values", "dynamic", "boolean_value"}
    task_df = pl.read_parquet(meds_dir / "tasks/boolean_value.parquet")
    code_index_df = pl.read_parquet(meds_dir / "triplet_tensors/metadata/codes.parquet")[
        "code", "code/vocab_index"
    ]
    target_indices = code_index_df.filter(
        pl.col("code").eq("ADMISSION//ONCOLOGY") | pl.col("code").eq("ADMISSION//CARDIAC")
    )["code/vocab_index"].to_list()
    items = []
    for index in range(len(pyd)):
        subject_data = pyd[index]
        subject_id, _, __loader__ = pyd.index[index]
        pyd.subj_map[subject_id]
        static_df_index = pyd.static_dfs[pyd.subj_map[subject_id]][pyd.subj_indices[subject_id]][
            "subject_id"
        ].item()
        # Check the subject ids match
        assert static_df_index == subject_id, f"Subject ids do not match for index {index}"

        # Check the supervised task matches the label
        assert (
            task_df.filter(pl.col("subject_id").eq(subject_id))[SUPERVISED_TASK_NAME].item()
            == subject_data[SUPERVISED_TASK_NAME]
        )
        assert subject_data[SUPERVISED_TASK_NAME] == pyd.labels[index]
        # Check the supervised task matches the target indices
        # The task is a deterministic function of the codes in this case
        # so we can check the codes themeselves are loaded correctly via this check.
        data_label = bool(
            functools.reduce(
                operator.or_, [subject_data["dynamic"].tensors["dim0/code"] == t for t in target_indices]
            ).any()
        )
        assert (
            data_label == subject_data[SUPERVISED_TASK_NAME]
        ), f"Supervised task does not match target indices for index {index}"

        items.append(subject_data)

    batch = pyd.collate(items)
    if collate_type in ["triplet", "triplet_prompt"]:
        assert batch.keys() == {
            "mask",
            "static_mask",
            "code",
            "numeric_value",
            "time_delta_days",
            "numeric_value_mask",
            SUPERVISED_TASK_NAME,
        }
        code_tensor = batch["code"]
        is_target_tensor = torch.zeros_like(code_tensor)
        for target_index in target_indices:
            is_target_tensor = is_target_tensor | code_tensor == target_index
        labels = [each[SUPERVISED_TASK_NAME] for each in items]
        # Check labels are in the same order after collating
        assert batch[SUPERVISED_TASK_NAME].to(bool).tolist() == labels
        # Data should have the right order after collating, using data derived labels
        data_label = functools.reduce(operator.or_, [code_tensor == t for t in target_indices]).any(axis=1)
        assert data_label.tolist() == labels, "Supervised task does not match target indices"
    elif collate_type == "eic":
        assert batch.keys() == {
            "mask",
            "static_mask",
            "code",
            "numeric_value",
            "time_delta_days",
            "numeric_value_mask",
            SUPERVISED_TASK_NAME,
        }
    else:
        raise NotImplementedError(f"{collate_type} not implemented")


@pytest.mark.parametrize("do_early_fuse_windows", [False, True])
@pytest.mark.parametrize("collate_type", ["triplet"])
@pytest.mark.parametrize("subject_level_sampling", [False, True])
def test_contrastive_windows(meds_dir, subject_level_sampling, collate_type, do_early_fuse_windows):
    overrides = [
        "data=multiwindow_pytorch_dataset",
        "data.do_include_subject_id=true",
    ]
    if do_early_fuse_windows:
        overrides.append("data.early_fusion_windows=[pre,post]")
    cfg = create_cfg(overrides=overrides, meds_dir=meds_dir)
    cfg.data.collate_type = collate_type
    cfg.data.subject_level_sampling = subject_level_sampling
    expected_item_keys = {"pre", "post", "subject_id"}

    assert cfg.data.cache_dir
    assert Path(cfg.data.raw_windows_fp).exists()

    pyd = MultiWindowPytorchDataset(cfg.data, split="train")
    item = pyd[0]
    assert item.keys() == expected_item_keys
    assert item["pre"].keys() == {"static_indices", "static_values", "dynamic"}
    assert item["post"].keys() == {"static_indices", "static_values", "dynamic"}

    expected_batch_keys = expected_item_keys | ({"FUSED"} if do_early_fuse_windows else set())
    batch = pyd.collate([pyd[i] for i in range(2)])
    assert batch.keys() == expected_batch_keys
    expected_window_keys = {
        "mask",
        "static_mask",
        "code",
        "numeric_value",
        "time_delta_days",
        "numeric_value_mask",
    }
    for window in expected_batch_keys - {"subject_id"}:
        assert batch[window].keys() & expected_window_keys == expected_window_keys


def test_full_datamodule(meds_dir):
    cfg = create_cfg(overrides=[], meds_dir=meds_dir)
    cfg.data.dataloader.batch_size = 1
    dm = MEDSDataModule(cfg.data)
    dm.prepare_data()

    assert not dm.data_train and not dm.data_val and not dm.data_test

    dm.setup()
    assert (
        isinstance(dm.data_train, PytorchDataset)
        and isinstance(dm.data_val, PytorchDataset)
        and isinstance(dm.data_test, PytorchDataset)
    )
    assert isinstance(dm.train_dataloader(), DataLoader) and isinstance(dm.val_dataloader(), DataLoader)

    num_datapoints = len(dm.data_train)
    assert num_datapoints == 70

    batch = next(iter(dm.train_dataloader()))
    for value in batch.values():
        assert len(value) == cfg.data.dataloader.batch_size


@pytest.mark.parametrize(
    "collate_type",
    [
        "triplet",
        "triplet_prompt",
        "eic",
    ],
)
def test_random_windows_pytorch_dataset(meds_dir, collate_type):
    cfg = create_cfg(overrides=["data=random_windows_pytorch_dataset"], meds_dir=meds_dir)
    with open_dict(cfg):
        cfg.data.collate_type = collate_type
        cfg.data.tokenizer = "emilyalsentzer/Bio_ClinicalBERT"
        cfg.data.min_window_size = 10
        cfg.data.max_window_size = 50
        cfg.data.n_windows = 1

    rwd = RandomWindowPytorchDataset(cfg.data, split="train")
    assert not rwd.has_task
    item = rwd[0]
    assert set(item.keys()) == {"window_0"}
    window = item["window_0"]
    assert set(window.keys()) == {"static_indices", "static_values", "dynamic"}

    batch = rwd.collate([rwd[i] for i in range(2)])
    assert set(batch.keys()) == {"window_0"}

    window = batch["window_0"]
    assert set(window.keys()) == {
        "mask",
        "static_mask",
        "code",
        "numeric_value",
        "time_delta_days",
        "numeric_value_mask",
    }


def test_random_window_generation(meds_dir):
    cfg = create_cfg(overrides=["data=random_windows_pytorch_dataset"], meds_dir=meds_dir)
    with open_dict(cfg):
        cfg.data.collate_type = "triplet"
        cfg.data.min_window_size = 10
        cfg.data.max_window_size = 50
        cfg.data.n_windows = 1

    rwd = RandomWindowPytorchDataset(
        cfg.data,
        split="train",
    )

    # Test generate_random_windows method
    windows = rwd.generate_random_windows(100)
    assert len(windows) == 1
    start, end = windows["window_0"]
    assert 10 <= end - start <= 50
    assert 0 <= start < end <= 100

    # Test that windows are different for different calls
    windows2 = rwd.generate_random_windows(100)
    assert windows["window_0"] != windows2["window_0"]


def test_random_window_batch_sizes(meds_dir):
    cfg = create_cfg(overrides=["data=random_windows_pytorch_dataset"], meds_dir=meds_dir)
    with open_dict(cfg):
        cfg.data.collate_type = "triplet"
        cfg.data.min_window_size = 10
        cfg.data.max_window_size = 50
        cfg.data.n_windows = 1

    rwd = RandomWindowPytorchDataset(cfg.data, split="train")

    batch = rwd.collate([rwd[i] for i in range(len(rwd))])
    assert len(batch["window_0"]["mask"]) == len(rwd)

    # Check that window sizes are within the specified range
    window_sizes = batch["window_0"]["mask"].sum(dim=1)
    assert all(10 <= size <= 50 for size in window_sizes)


def test_random_window_generation_modes(meds_dir):
    cfg = create_cfg(overrides=["data=random_windows_pytorch_dataset"], meds_dir=meds_dir)
    with open_dict(cfg):
        cfg.data.collate_type = "triplet"
        cfg.data.n_windows = 3
        cfg.data.window_names = ["window_1", "window_2", "window_3"]
        cfg.data.max_window_size = 50

    # Test consecutive windows
    with open_dict(cfg):
        cfg.data.consecutive_windows = True

    rwd_consecutive = RandomWindowPytorchDataset(cfg.data, split="train")

    # Generate windows for a hypothetical sequence of length 200
    consecutive_windows = rwd_consecutive.generate_random_windows(200)

    # Check if windows are consecutive
    window_starts = [w[0] for w in consecutive_windows.values()]
    window_ends = [w[1] for w in consecutive_windows.values()]

    assert all(
        window_ends[i] == window_starts[i + 1] for i in range(len(window_starts) - 1)
    ), "Consecutive windows are not truly consecutive"

    assert window_starts == sorted(window_starts), "Consecutive windows are not in order"

    # Test non-consecutive windows
    with open_dict(cfg):
        cfg.data.consecutive_windows = False

    rwd_non_consecutive = RandomWindowPytorchDataset(cfg.data, split="train")

    # Generate windows for a hypothetical sequence of length 200
    non_consecutive_windows = rwd_non_consecutive.generate_random_windows(200)

    # Check if windows are non-consecutive (at least one pair is non-consecutive)
    window_starts = [w[0] for w in non_consecutive_windows.values()]
    window_ends = [w[1] for w in non_consecutive_windows.values()]

    assert any(
        window_ends[i] != window_starts[i + 1] for i in range(len(window_starts) - 1)
    ), "Non-consecutive windows are all consecutive"

    # Additional checks
    for mode, windows in [("Consecutive", consecutive_windows), ("Non-consecutive", non_consecutive_windows)]:
        # Check if all windows are within the sequence length
        assert all(
            0 <= start < end <= 200 for start, end in windows.values()
        ), f"{mode} windows exceed sequence length"

        # Check if window sizes are correct
        assert all(
            0 < end - start <= cfg.data.max_window_size for start, end in windows.values()
        ), f"{mode} window sizes are incorrect"

        # Check if the correct number of windows is generated
        assert len(windows) == cfg.data.n_windows, f"Incorrect number of {mode.lower()} windows generated"

        # Check if window names are correct
        assert set(windows.keys()) == set(
            cfg.data.window_names
        ), f"Incorrect window names for {mode.lower()} windows"

    print("All window generation tests passed successfully!")



import lightning as L
import torch
from omegaconf import DictConfig

from meds_torch.utils.module_class import Module


class BaseModule(L.LightningModule, Module):
    def __init__(
        self,
        cfg: DictConfig,
    ):
        """Initializes the BaseModule with the given configuration, setting up various components such as the
        optimizer, scheduler, model, and input encoder.

        Parameters: cfg (DictConfig): The configuration dictionary specifying the setup of the module,
        including required elements like the task name and optional elements                   for configuring
        the optimizer, scheduler, model, and input encoder.

        Raises: ValueError: If the task name is not specified in the configuration.

        Returns: None
        """
        super().__init__()
        self.cfg = cfg
        # shared components
        self.optimizer = cfg.optimizer
        self.scheduler = cfg.get("scheduler", None)
        self.model = cfg.backbone
        self.input_encoder = cfg.input_encoder

    def configure_optimizers(self):
        optimizer = self.optimizer(self.parameters())
        if self.scheduler is not None:
            scheduler = self.scheduler.instantiate(optimizer)
            return {
                "optimizer": optimizer,
                "lr_scheduler": {"scheduler": scheduler, **self.scheduler.extra_kwargs},
            }
        return optimizer

    def setup(self, stage: str) -> None:
        """Lightning hook that is called at the beginning of fit (train + validate), validate, test, or
        predict.

        This is a good hook when you need to build models dynamically or adjust something about them. This
        hook is called on every process when using DDP.

        :param stage: Either `"fit"`, `"validate"`, `"test"`, or `"predict"`.
        """
        if self.cfg.compile and stage == "fit":
            self.net = torch.compile(self.net)



"""Adapted from:

https://github.com/PyTorchLightning/pytorch-lightning/blob/master/tests/helpers/runif.py
"""

import sys
from typing import Any

import pytest
import torch
from packaging.version import Version
from pkg_resources import get_distribution
from pytest import MarkDecorator

from tests.helpers.package_available import (
    _COMET_AVAILABLE,
    _DEEPSPEED_AVAILABLE,
    _DO_LOG,
    _FAIRSCALE_AVAILABLE,
    _IS_WINDOWS,
    _MLFLOW_AVAILABLE,
    _NEPTUNE_AVAILABLE,
    _SH_AVAILABLE,
    _TPU_AVAILABLE,
    _WANDB_AVAILABLE,
)


class RunIf:
    """RunIf wrapper for conditional skipping of tests.

    Fully compatible with `@pytest.mark`.

    Example:

    ```python     @RunIf(min_torch="1.8")     @pytest.mark.parametrize("arg1", [1.0, 2.0])     def
    test_wrapper(arg1):         assert arg1 > 0 ```
    """

    def __new__(
        cls,
        min_gpus: int = 0,
        min_torch: str | None = None,
        max_torch: str | None = None,
        min_python: str | None = None,
        skip_windows: bool = False,
        sh: bool = False,
        tpu: bool = False,
        fairscale: bool = False,
        deepspeed: bool = False,
        wandb: bool = False,
        neptune: bool = False,
        comet: bool = False,
        mlflow: bool = False,
        do_log: bool = False,
        **kwargs: dict[Any, Any],
    ) -> MarkDecorator:
        """Creates a new `@RunIf` `MarkDecorator` decorator.

        :param min_gpus: Min number of GPUs required to run test.
        :param min_torch: Minimum pytorch version to run test.
        :param max_torch: Maximum pytorch version to run test.
        :param min_python: Minimum python version required to run test.
        :param skip_windows: Skip test for Windows platform.
        :param tpu: If TPU is available.
        :param sh: If `sh` module is required to run the test.
        :param fairscale: If `fairscale` module is required to run the test.
        :param deepspeed: If `deepspeed` module is required to run the test.
        :param wandb: If `wandb` module is required to run the test.
        :param neptune: If `neptune` module is required to run the test.
        :param comet: If `comet` module is required to run the test.
        :param mlflow: If `mlflow` module is required to run the test.
        :param kwargs: Native `pytest.mark.skipif` keyword arguments.
        """
        conditions = []
        reasons = []

        if min_gpus:
            conditions.append(torch.cuda.device_count() < min_gpus)
            reasons.append(f"GPUs>={min_gpus}")

        if min_torch:
            torch_version = get_distribution("torch").version
            conditions.append(Version(torch_version) < Version(min_torch))
            reasons.append(f"torch>={min_torch}")

        if max_torch:
            torch_version = get_distribution("torch").version
            conditions.append(Version(torch_version) >= Version(max_torch))
            reasons.append(f"torch<{max_torch}")

        if min_python:
            py_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
            conditions.append(Version(py_version) < Version(min_python))
            reasons.append(f"python>={min_python}")

        if skip_windows:
            conditions.append(_IS_WINDOWS)
            reasons.append("does not run on Windows")

        if tpu:
            conditions.append(not _TPU_AVAILABLE)
            reasons.append("TPU")

        if sh:
            conditions.append(not _SH_AVAILABLE)
            reasons.append("sh")

        if fairscale:
            conditions.append(not _FAIRSCALE_AVAILABLE)
            reasons.append("fairscale")

        if deepspeed:
            conditions.append(not _DEEPSPEED_AVAILABLE)
            reasons.append("deepspeed")

        if wandb:
            conditions.append(not _WANDB_AVAILABLE)
            reasons.append("wandb")

        if neptune:
            conditions.append(not _NEPTUNE_AVAILABLE)
            reasons.append("neptune")

        if comet:
            conditions.append(not _COMET_AVAILABLE)
            reasons.append("comet")

        if mlflow:
            conditions.append(not _MLFLOW_AVAILABLE)
            reasons.append("mlflow")

        if do_log:
            conditions.append(not _DO_LOG)
            reasons.append("Logging is disabled")

        reasons = [rs for cond, rs in zip(conditions, reasons) if cond]
        return pytest.mark.skipif(
            condition=any(conditions),
            reason=f"Requires: [{' + '.join(reasons)}]",
            **kwargs,
        )



import dataclasses
import enum

import torch
from omegaconf import DictConfig
from torch import nn

from meds_torch.input_encoder import INPUT_ENCODER_MASK_KEY, INPUT_ENCODER_TOKENS_KEY
from meds_torch.utils.module_class import Module


@dataclasses.dataclass
class ModelOutput:
    rep: torch.Tensor
    hidden_states: torch.Tensor = None


class Triplet(enum.Enum):
    DATE = "date"
    VALUE = "value"
    VARIABLE = "variable"


def sequence_mask(lengths, maxlen, dtype=torch.bool):
    row_vector = torch.arange(0, maxlen, 1, device=lengths.device)
    matrix = torch.unsqueeze(lengths, dim=-1)
    mask = row_vector < matrix

    mask.type(dtype)
    return mask


class CVE(nn.Module):
    """Continuous Value Encoder (CVE) module.

    Assumes input is a single continuous value, and encodes it as an `output_dim` size embedding vector.
    """

    def __init__(self, cfg):
        super().__init__()
        self.layer = nn.Linear(1, cfg.token_dim)

    def forward(self, x):
        return self.layer(x)


class TripletPromptEncoder(nn.Module, Module):
    """Container module with an encoder, a recurrent or transformer module, and a decoder.

    Copied from: https://github.com/pytorch/examples/blob/main/word_language_model/model.py
    """

    def __init__(self, cfg: DictConfig):
        super().__init__()
        self.cfg = cfg
        # Define Triplet Embedders
        # TODO add to config
        self.code_embedder = torch.nn.Embedding(cfg.vocab_size + 2, embedding_dim=cfg.token_dim)
        self.numeric_value_embedder = CVE(cfg)

    def embed_func(self, embedder, x):
        out = embedder.forward(x[None, :].transpose(2, 0)).permute(1, 2, 0)
        return out

    def get_embedding(self, batch):
        code = batch["code"]
        numeric_value = batch["numeric_value"]
        numeric_value_mask = batch["numeric_value_mask"]

        # Embed codes
        code_emb = self.code_embedder.forward(code).permute(0, 2, 1)
        # Embed numerical values and mask nan values
        val_emb = self.embed_func(self.numeric_value_embedder, numeric_value) * numeric_value_mask.unsqueeze(
            dim=1
        )

        # Sum the (time, code, value) triplets and
        embedding = code_emb + val_emb

        assert embedding.isfinite().all(), "Embedding is not finite"
        return embedding

    def forward(self, batch):
        embedding = self.get_embedding(batch)
        batch[INPUT_ENCODER_MASK_KEY] = batch["mask"]
        batch[INPUT_ENCODER_TOKENS_KEY] = embedding
        return batch



"""Transformer Encoder Model.

Uses transformer encoder to get contextualized representations of all tokens and we take the CLS token
representation as the embedding.
"""

import torch
from omegaconf import DictConfig

from meds_torch.input_encoder import INPUT_ENCODER_MASK_KEY, INPUT_ENCODER_TOKENS_KEY
from meds_torch.models import BACKBONE_EMBEDDINGS_KEY, BACKBONE_TOKENS_KEY
from meds_torch.utils.module_class import Module


class TransformerEncoderModel(torch.nn.Module, Module):
    """Wrapper of Encoder Transformer for use in MEDS with triplet token embeddings."""

    def __init__(self, cfg: DictConfig):
        super().__init__()
        self.model = cfg.model

        if cfg.token_emb:
            self.model.token_emb = cfg.token_emb

    def forward(self, batch):
        input_data, mask = batch[INPUT_ENCODER_TOKENS_KEY], batch[INPUT_ENCODER_MASK_KEY]
        output, embeddings = self.model(input_data, mask=mask, return_logits_and_embeddings=True)
        batch[BACKBONE_TOKENS_KEY] = output
        batch[BACKBONE_EMBEDDINGS_KEY] = embeddings
        return batch



import shutil
from importlib.resources import files
from typing import Any

import hydra
import lightning as L
import loguru
from lightning import Callback, LightningDataModule, LightningModule, Trainer
from lightning.pytorch.loggers import Logger
from omegaconf import DictConfig

from meds_torch.utils import (
    RankedLogger,
    configure_logging,
    get_metric_value,
    instantiate_callbacks,
    instantiate_loggers,
    log_hyperparameters,
    task_wrapper,
)
from meds_torch.utils.resolvers import setup_resolvers

setup_resolvers()
log = RankedLogger(__name__, rank_zero_only=True)
config_yaml = files("meds_torch").joinpath("configs/train.yaml")


def initialize_train_objects(cfg: DictConfig, **kwargs) -> Trainer:
    """Instantiates a Lightning Trainer object.

    :param cfg: A DictConfig configuration composed by Hydra.
    :return: A Lightning Trainer object.
    """
    # set seed for random number generators in pytorch, numpy and python.random
    if cfg.get("seed"):
        L.seed_everything(cfg.seed, workers=True)

    log.info(f"Instantiating datamodule <{cfg.data._target_}>")
    datamodule: LightningDataModule = hydra.utils.instantiate(cfg.data)

    log.info(f"Instantiating model <{cfg.model._target_}>")
    model: LightningModule = hydra.utils.instantiate(cfg.model)

    log.info("Instantiating callbacks...")
    callbacks: list[Callback] = instantiate_callbacks(cfg.get("callbacks"))

    log.info("Instantiating loggers...")
    logger: list[Logger] = instantiate_loggers(cfg.get("logger"))

    log.info(f"Instantiating trainer <{cfg.trainer._target_}>")
    trainer_factory: Trainer = hydra.utils.instantiate(
        cfg.trainer, callbacks=callbacks, logger=logger, _partial_=True
    )
    trainer = trainer_factory(**kwargs)

    object_dict = {
        "cfg": cfg,
        "datamodule": datamodule,
        "model": model,
        "callbacks": callbacks,
        "logger": logger,
        "trainer": trainer,
    }

    return object_dict


@task_wrapper
def train(cfg: DictConfig) -> tuple[dict[str, Any], dict[str, Any]]:
    """Trains the model. Can additionally evaluate on a testset, using best weights
    obtained during training.

    This method is wrapped in optional @task_wrapper decorator, that controls the
    behavior during failure. Useful for multiruns, saving info about the crash, etc.

    Args:
        cfg: A DictConfig configuration composed by Hydra.
    Returns:
        A tuple with metrics and dict with all instantiated objects.
    """
    # cache hydra config

    object_dict = initialize_train_objects(cfg)
    logger = object_dict["logger"]
    trainer = object_dict["trainer"]
    model = object_dict["model"]
    datamodule = object_dict["datamodule"]

    if logger:
        log.info("Logging hyperparameters!")
        log_hyperparameters(object_dict)

    if cfg.get("train"):
        log.info("Starting training!")
        trainer.fit(model=model, datamodule=datamodule, ckpt_path=cfg.get("ckpt_path"))

    train_metrics = trainer.callback_metrics

    if cfg.get("test"):
        log.info("Starting testing!")
        best_model_path = trainer.checkpoint_callback.best_model_path
        if "strategy" in cfg.trainer and "ddp" in cfg.trainer.strategy:
            metric_dict = {**train_metrics}
            return metric_dict, object_dict, best_model_path

        if best_model_path == "":
            log.warning("Best ckpt not found! Using current weights for testing...")
            best_model_path = None
        trainer.test(model=model, datamodule=datamodule, ckpt_path=best_model_path)
        log.info(f"Best ckpt path: {best_model_path}")

    test_metrics = trainer.callback_metrics

    # merge train and test metrics
    metric_dict = {**train_metrics, **test_metrics}

    # log time profiles: https://github.com/Oufattole/meds-torch/issues/44
    if hasattr(datamodule.data_train, "_timings"):
        loguru.logger.debug("Train Time: ", datamodule.data_train._profile_durations())

    return metric_dict, object_dict, best_model_path


@hydra.main(
    version_base="1.3",
    config_path=str(config_yaml.parent.resolve()),
    config_name=config_yaml.stem,
)
def main(cfg: DictConfig) -> float | None:
    """Main entry point for training.

    Args:
        cfg: DictConfig configuration composed by Hydra.
    Returns:
        Optional[float] with optimized metric value.
    """
    # apply extra utilities
    # (e.g. ask for tags if none are provided in cfg, print cfg tree, etc.)
    configure_logging(cfg)

    # train the model
    metric_dict, _, best_model_path = train(cfg)

    # safely retrieve metric value for hydra-based hyperparameter optimization
    metric_value = get_metric_value(metric_dict=metric_dict, metric_name=cfg.get("optimized_metric"))

    checkpoint_dir = cfg.paths.time_output_dir / "checkpoints"
    checkpoint_dir.mkdir(parents=True, exist_ok=True)

    if not cfg.trainer.get("fast_dev_run"):
        shutil.copy(best_model_path, checkpoint_dir / "best_model.ckpt")

    # return optimized metric
    return metric_value


if __name__ == "__main__":
    main()



import numpy as np
import polars as pl
import torch
import torch.nn.functional as F
import torch.utils
from clinical_zeroshot_labeler.labeler import SequenceLabeler, WindowStatus
from clinical_zeroshot_labeler.model import BaseGenerativeModel
from loguru import logger
from mixins import TimeableMixin
from omegaconf import DictConfig
from torchmetrics import Metric, MetricCollection
from torchmetrics.classification import MulticlassAccuracy, MulticlassAUROC
from x_transformers import Decoder, TransformerWrapper
from x_transformers.autoregressive_wrapper import eval_decorator

from meds_torch.input_encoder import INPUT_ENCODER_MASK_KEY, INPUT_ENCODER_TOKENS_KEY
from meds_torch.models import (
    BACKBONE_EMBEDDINGS_KEY,
    BACKBONE_TOKENS_KEY,
    GENERATE_PREFIX,
    MODEL_BATCH_LOSS_KEY,
    MODEL_EMBEDDINGS_KEY,
    MODEL_LOGITS_SEQUENCE_KEY,
    MODEL_LOSS_KEY,
    MODEL_PRED_PROBA_KEY,
    MODEL_PRED_STATUS_KEY,
    MODEL_PREFIX,
    MODEL_TOKENS_KEY,
)
from meds_torch.models.base_model import BaseModule
from meds_torch.models.components.utils import TrajectoryBatch, get_time_days_delta
from meds_torch.utils.custom_time_token import TIME_DELTA_TOKEN


class DummyTrajectoryLabeler:
    def __init__(self, B):
        self.counter = 0
        self.status = [
            torch.tensor([WindowStatus.UNDETERMINED.value] * B),
            torch.tensor([WindowStatus.ACTIVE.value] * B),
            torch.tensor([WindowStatus.SATISFIED.value] * B),
        ]
        self.labels = torch.zeros((B,), dtype=torch.bool)

    def process_step(self, tokens, times, values):
        status = self.status[min(self.counter, len(self.status) - 1)]
        self.counter += 1
        return status

    def is_finished(self):
        return self.counter >= len(self.status)

    def get_labels(self):
        return self.labels


def create_dummy_sequence_labeler(batch_size: int = 2):
    """Create a dummy sequence labeler with a simple ACES task configuration.

    Args:
        batch_size: Number of sequences to process in parallel

    Returns:
        Tuple containing:
            - Dummy labeler instance
            - Metadata DataFrame
            - Sample input batch
            - ACES task configuration string

    Examples:
        >>> labeler, metadata_df, batch, task_config = create_dummy_sequence_labeler()
        >>> import torch
        >>> assert isinstance(batch['code'], torch.Tensor)
        >>> assert batch['code'].shape == (2, 3)  # batch_size=2, seq_len=3
        >>> assert 'mask' in batch
        >>> # Test indices are within vocab range
        >>> max_idx = batch['code'].max()
        >>> assert max_idx < len(metadata_df)
    """
    from datetime import datetime

    import polars as pl
    import torch
    from clinical_zeroshot_labeler.labeler import SequenceLabeler

    # Define simple ACES task configuration
    task_config = """
    predicates:
        hospital_discharge:
            code: {regex: "HOSPITAL_DISCHARGE//.*"}
        lab:
            code: {regex: "LAB//.*"}
        high_lab:
            code: {regex: "LAB//.*"}
            value_min: 2.0
            value_min_inclusive: True

    trigger: hospital_discharge

    windows:
        input:
            start: NULL
            end: trigger
            start_inclusive: True
            end_inclusive: True
            index_timestamp: end
        target:
            start: input.end
            end: start + 365d
            start_inclusive: False
            end_inclusive: True
            has:
                lab: (1, None)
            label: high_lab
    """

    # Create metadata DataFrame with test codes
    metadata_df = pl.DataFrame(
        {
            "code": [
                "PAD",
                "HOSPITAL_DISCHARGE//MEDICAL",
                "LAB//_Q_1",
                "LAB//_Q_2",
                "LAB//_Q_3",
                "TIME//DELTA//TOKEN//_Q_1",
                "TIME//DELTA//TOKEN//_Q_2",
                "TIME//DELTA//TOKEN//_Q_3",
            ],
            "code/vocab_index": [0, 1, 2, 3, 4, 5, 6, 7],
            "values/min": [None, None, 2.0, 0.0, 1.0, 0, 1, 2],
            "values/max": [None, None, 3.0, 1.0, 2.0, 1, 2, 3],
            "values/sum": [None, None, 0.5, 1.5, 2.5, 0.5, 1.5, 2.5],
            "values/n_occurrences": [None, None, 1, 1, 1, 1, 1, 1],
            "values/quantiles": [
                {"values/quantile/0.5": None},
                {"values/quantile/0.5": None},
                {"values/quantile/0.5": 1},
                {"values/quantile/0.5": 1},
                {"values/quantile/0.5": 1},
                {"values/quantile/0.5": 1},
                {"values/quantile/0.5": 1},
                {"values/quantile/0.5": 1},
            ],
        }
    )

    # Create sample input batch
    # Sequence 1: Hospital discharge -> High lab value -> Time token
    # Sequence 2: Hospital discharge -> Normal lab value -> Time token
    # Note: All indices should be < len(metadata_df)
    batch = {
        "code": torch.tensor([[1, 2, 4], [1, 3, 4]]),  # Using vocab indices
        "mask": torch.ones(2, 3, dtype=torch.bool),
        "subject_id": torch.tensor([1, 2]),
        "prediction_time": [datetime(2020, 1, 1), datetime(2020, 1, 1)],
        "end_time": [datetime(2020, 1, 1), datetime(2020, 1, 1)],
    }

    # Initialize sequence labeler
    from functools import partial

    labeler = partial(SequenceLabeler.from_yaml_str, yaml_str=task_config, early_stop=True)

    return labeler, metadata_df, batch, task_config


def create_model_config(metadata_df_path: str):
    """Create a model configuration for testing.

    Args:
        metadata_df_path: Path to metadata DataFrame parquet file

    Returns:
        Instantiated model configuration

    Examples:
        >>> import tempfile, polars as pl
        >>> with tempfile.NamedTemporaryFile(suffix='.parquet') as temp_file:
        ...     df = pl.DataFrame({"code": ["A"], "code/vocab_index": [0]})
        ...     df.write_parquet(temp_file.name)
        ...     cfg = create_model_config(temp_file.name)
        >>> assert cfg.vocab_size == 2  # Original size + pad token
    """
    from hydra.utils import instantiate

    vocab_size = pl.read_parquet(metadata_df_path).height + 1
    cfg = {
        "code_metadata_fp": metadata_df_path,
        "backbone": {"_target_": "meds_torch.models.eic_forecasting.DummyModel"},
        "vocab_size": vocab_size,  # Add 1 for pad token
        "generate_id": None,
        "store_generated_trajectory": True,
        "max_seq_len": 10,
        "temperature": 1.0,
        "eos_tokens": None,
        "optimizer": {"_target_": "meds_torch.models.eic_forecasting.DummyOptimizer", "_partial_": True},
        "scheduler": {"_target_": "meds_torch.models.eic_forecasting.DummyScheduler", "_partial_": True},
        "input_encoder": {"_target_": "meds_torch.models.eic_forecasting.DummyEncoder"},
        "code_head": {
            "_target_": "meds_torch.models.eic_forecasting.DummyCodeHead",
            "vocab_size": vocab_size,
        },
        "compile": False,
        "top_k_acc": [1],
        "next_token_auc": False,
        "max_tokens_budget": 10,
        "return_tokens": False,
        "return_logits": False,
        "return_labeler": False,
        "prune_terminated": False,
    }
    return instantiate(cfg)


class DummyModel:
    """Dummy model that generates two fixed sequences."""

    cfg = DictConfig(dict(token_emb=None))

    def __init__(self):
        self.model = TransformerWrapper(
            num_tokens=5,
            max_seq_len=10,
            attn_layers=Decoder(dim=8, depth=1, heads=2, rotary_pos_emb=True),
            use_abs_pos_emb=False,
        )

    def __call__(self, batch):
        B, S = batch["code"].shape
        return {BACKBONE_TOKENS_KEY: torch.ones(B, S, 32), BACKBONE_EMBEDDINGS_KEY: None}

    def generate(self, prompts, **kwargs):
        # Always generate two fixed sequences
        generated = torch.tensor(
            [
                [5, 2, 2, 7, 7],  # Sequence 1: low labs only
                [5, 4, 4, 5, 5],  # Sequence 2: high labs only
            ]
        )
        out_lengths = torch.tensor([5, 5])
        labels = dict()
        if kwargs.get("trajectory_labeler") is not None:
            labels = dict(
                labels=torch.tensor([1.0, 0.0]),  # Sequence 1 positive, Sequence 2 negative
                status=torch.ones(2) * WindowStatus.SATISFIED.value,
            )
        return generated, out_lengths, labels


class DummyCodeHead:
    def __init__(self, vocab_size):
        self.vocab_size = vocab_size  # Match metadata size

    def __call__(self, x):
        B, S, _ = x.shape
        logits = torch.zeros(B, S, self.vocab_size)
        logits[..., 1:] = 1.0  # All tokens except PAD equally likely
        return logits


class DummyEncoder:
    def __call__(self, batch):
        batch[INPUT_ENCODER_TOKENS_KEY] = batch["code"]
        batch[INPUT_ENCODER_MASK_KEY] = torch.ones_like(batch["mask"]).bool()
        return batch


class DummyOptimizer:
    def __init__(self, params):
        self.params = list(params)

    def step(self):
        pass

    def zero_grad(self):
        pass


class DummyScheduler:
    def step(self):
        pass

    def get_last_lr(self):
        return [0.001]


CODE_LOGITS = "EIC_MODEL//CODE_LOGITS"


# Function to pad a single array
def pad_array(arr, max_len):
    pad_width = ((0, 0), (0, max_len - arr.shape[1]))
    if arr.dtype == bool:
        return np.pad(arr, pad_width, mode="constant", constant_values=False)
    else:
        return np.pad(arr, pad_width, mode="constant", constant_values=0)


class NextTokenPredictionMetric(Metric):
    """
    A metric class for calculating AUC and top-n accuracy for next token prediction in language models.

    This metric computes the Area Under the Receiver Operating Characteristic Curve (AUROC) and
    top-n accuracy for each position in the sequence, considering only the next token prediction.

    Attributes:
        vocab_size (int): The size of the vocabulary.
        top_n (tuple): The values of n for which to calculate top-n accuracy.
        auroc (MulticlassAUROC): The AUROC metric for multiclass classification.
        top_n_accuracy (dict): A dictionary of MulticlassAccuracy metrics for each n in top_n.
    """

    def __init__(self, vocab_size: int, top_k_acc: list[int], next_token_auc: bool, dist_sync_on_step=False):
        """
        Initialize the NextTokenPredictionMetric.

        Args:
            vocab_size (int): The size of the vocabulary.
            top_n (tuple): The values of n for which to calculate top-n accuracy. Default is (1, 5, 10).
            dist_sync_on_step (bool): Synchronize metric state across processes at each step. Default is
                False.
        """
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.vocab_size = vocab_size

        self.top_k_acc = top_k_acc
        metrics = {
            f"top_{k}_accuracy": MulticlassAccuracy(num_classes=vocab_size, top_k=k) for k in top_k_acc
        }
        if next_token_auc:
            metrics["auroc"] = MulticlassAUROC(num_classes=vocab_size, average="macro", thresholds=100)
        self.next_token_metrics = MetricCollection(metrics)

    def update(self, logits: torch.Tensor, targets: torch.Tensor, mask: torch.Tensor):
        """
        Update the metric state with batch statistics.

        Args:
            logits (torch.Tensor): Predicted logits from the model, shape (batch_size, seq_length,
                vocab_size).
            targets (torch.Tensor): Ground truth labels, shape (batch_size, seq_length).
            mask (torch.Tensor): Mask to ignore padded elements, shape (batch_size,
                seq_length).

        The method shifts the targets to align with the next token prediction and updates AUROC and top-n
            accuracy.
        """

        # Shift targets to align with next token prediction
        shifted_targets = targets[:, 1:]
        shifted_mask = mask[:, :-1]

        # Reshape tensors for metric update
        flat_logits = logits[:, :-1][shifted_mask].view(-1, self.vocab_size)
        flat_targets = shifted_targets[shifted_mask].view(-1)

        # Update metrics
        self.next_token_metrics.update(flat_logits, flat_targets)

    def compute(self):
        """
        Compute the AUROC and top-n accuracy based on accumulated statistics.

        Returns:
            dict: A dictionary containing the computed AUROC and top-n accuracy for each n in top_n.
        """
        results = self.next_token_metrics.compute()
        return results


class EicForecastingModule(BaseModule, TimeableMixin, BaseGenerativeModel):
    """EIC token based GPT Forecasting Model.

    This model has three main capabilities:
    1. Autoregressive training (learning to predict next tokens)
    2. Data generation (creating synthetic medical event sequences)
    3. Zero-shot prediction (using generated sequences for prediction)

    Args:
        cfg (DictConfig): Configuration object containing:
            - vocab_size: Size of the vocabulary
            - max_seq_len: Maximum sequence length
            - zero_shot_labeler: Optional function for zero-shot prediction
            - code_metadata_fp: Path to code metadata file

    Examples:
        >>> import tempfile
        >>> from clinical_zeroshot_labeler.labeler import WindowStatus
        >>> # Create test setup using helper function
        >>> trajectory_labeler, metadata_df, batch, _ = create_dummy_sequence_labeler()

        >>> # Write metadata to temporary file and create config
        >>> temp_file = tempfile.NamedTemporaryFile(suffix='.parquet')
        >>> metadata_df.write_parquet(temp_file.name)
        >>> cfg = create_model_config(temp_file.name)

        >>> # Test workflow 1: Autoregressive training
        >>> model = EicForecastingModule(cfg)
        >>> loss = model.training_step(batch)
        >>> assert loss.isfinite().all()

        >>> # Test workflow 2: Data generation without labeling
        >>> cfg.generate_id = 1
        >>> model = EicForecastingModule(cfg)
        >>> output = model.forward(batch)
        >>> assert GENERATE_PREFIX + '1' in output
        >>> generated_df = output[GENERATE_PREFIX + '1']
        >>> # Check generated data structure
        >>> assert 'time' in generated_df.columns
        >>> assert 'code' in generated_df.columns
        >>> assert 'numeric_value' in generated_df.columns
        >>> assert 'subject_id' in generated_df.columns
        >>> assert 'prediction_time' in generated_df.columns
        >>> # Verify time token generation (code/vocab_index 4 in metadata)
        >>> generated_df.shape[0]
        20

        >>> # Test workflow 3: Generation with zero-shot labeling
        >>> cfg.generate_id = 1
        >>> model = EicForecastingModule(cfg)
        >>> model.trajectory_labeler = trajectory_labeler
        >>> output = model.forward(batch)
        >>> # Check labeling output
        >>> assert MODEL_PRED_PROBA_KEY in output
        >>> assert MODEL_PRED_STATUS_KEY in output
        >>> assert output[MODEL_PRED_PROBA_KEY].shape == (2,)  # Binary prediction per sequence
        >>> assert output[MODEL_PRED_STATUS_KEY].shape == (2,)  # Status per sequence
        >>> # Verify status progression works
        >>> status_vals = output[MODEL_PRED_STATUS_KEY]
        >>> assert (status_vals == WindowStatus.SATISFIED.value).any(), status_vals  # Some sequences complete
    """

    def __init__(self, cfg: DictConfig):
        super().__init__(cfg)
        self.code_head = self.cfg.code_head

        num_future_codes = self.cfg.get("num_future_codes", None)
        if num_future_codes is not None:
            logger.info(f"Using {num_future_codes} future codes for forecasting")
        self.train_next_token_metric = NextTokenPredictionMetric(
            self.cfg.vocab_size, self.cfg.top_k_acc, self.cfg.next_token_auc
        )
        self.val_next_token_metric = NextTokenPredictionMetric(
            self.cfg.vocab_size, self.cfg.top_k_acc, self.cfg.next_token_auc
        )
        self.test_next_token_metric = NextTokenPredictionMetric(
            self.cfg.vocab_size, self.cfg.top_k_acc, self.cfg.next_token_auc
        )

        self.metadata_df = pl.read_parquet(self.cfg.code_metadata_fp)
        self.trajectory_labeler = self.cfg.get("trajectory_labeler", None)

    def get_loss(self, batch):
        code_logits = batch[CODE_LOGITS]
        assert not torch.isnan(code_logits).any(), "code_logits is NaN"

        # Code Mask
        mask = batch["mask"]
        code_target = batch["code"]

        # Shift the target to predict the next token
        shifted_code_target = code_target[:, 1:]  # Remove the first token
        shifted_mask = mask[:, 1:]  # Remove the first position from the mask too

        # Apply the mask to code_logits and shifted_code_target
        masked_code_logits = code_logits[:, :-1] * shifted_mask.unsqueeze(-1)  # Remove the last prediction
        masked_code_target = shifted_code_target * shifted_mask

        # Code Loss
        code_loss = F.cross_entropy(
            masked_code_logits.transpose(1, 2),
            masked_code_target.to(dtype=torch.long),
            reduction="none",
        ).mean(dim=-1)

        assert not torch.isnan(code_loss).any(), "code_loss is NaN"

        return code_loss

    def get_forecast_logits(self, model_output):
        if isinstance(model_output, torch.Tensor):
            all_token_embeddings = model_output
        else:
            all_token_embeddings = model_output[BACKBONE_TOKENS_KEY]
        code_logits = self.code_head(all_token_embeddings)
        return {
            CODE_LOGITS: code_logits,
        }

    def forward(self, batch, keep_code_logits=False):
        batch = self.input_encoder(batch)
        model_output = self.model(batch)

        if self.cfg.return_tokens:
            batch[MODEL_TOKENS_KEY] = model_output[BACKBONE_TOKENS_KEY]
        batch[MODEL_EMBEDDINGS_KEY] = model_output[BACKBONE_EMBEDDINGS_KEY]
        forecast = self.get_forecast_logits(model_output)
        if self.cfg.return_logits:
            batch[MODEL_LOGITS_SEQUENCE_KEY] = forecast[CODE_LOGITS]
        batch[CODE_LOGITS] = forecast[CODE_LOGITS]

        code_loss = self.get_loss(batch)
        batch[MODEL_LOSS_KEY] = code_loss
        batch[MODEL_BATCH_LOSS_KEY] = code_loss.mean()
        batch = self._generate(batch)

        if not keep_code_logits:
            del batch[CODE_LOGITS]
        return batch

    def _log(self, batch, split):
        self.log(split + "/loss", batch[MODEL_BATCH_LOSS_KEY])
        if split == "train":
            self.train_next_token_metric.update(batch[CODE_LOGITS], batch["code"], batch["mask"])
        elif split == "val":
            self.val_next_token_metric.update(batch[CODE_LOGITS], batch["code"], batch["mask"])
        elif split == "test":
            self.test_next_token_metric.update(batch[CODE_LOGITS], batch["code"], batch["mask"])

    def _generate(self, batch):
        if self.cfg.generate_id is not None:
            return self.generate_batch(batch)
        else:
            return batch

    def training_step(self, batch):
        batch = self(batch, True)
        assert not torch.isnan(batch[MODEL_BATCH_LOSS_KEY]), "Loss is NaN"
        self._log(batch, "train")
        del batch[CODE_LOGITS]
        return batch[MODEL_BATCH_LOSS_KEY]

    def on_train_epoch_end(self):
        next_token_results = self.train_next_token_metric.compute()
        for metric_name, value in next_token_results.items():
            self.log(f"test/NEXT_TOKEN/{metric_name.upper()}", value, on_epoch=True)
        self.train_next_token_metric.reset()

    def validation_step(self, batch):
        batch = self(batch, True)
        assert not torch.isnan(batch[MODEL_BATCH_LOSS_KEY]), "Loss is NaN"
        self._log(batch, "val")
        del batch[CODE_LOGITS]
        return batch[MODEL_BATCH_LOSS_KEY]

    def on_validation_epoch_end(self):
        next_token_results = self.val_next_token_metric.compute()
        for metric_name, value in next_token_results.items():
            self.log(f"test/NEXT_TOKEN/{metric_name.upper()}", value, on_epoch=True)
        self.val_next_token_metric.reset()

    def test_step(self, batch):
        batch = self(batch, True)
        assert not torch.isnan(batch[MODEL_BATCH_LOSS_KEY]), "Loss is NaN"
        self._log(batch, "test")
        del batch[CODE_LOGITS]
        loss = batch[MODEL_BATCH_LOSS_KEY]
        return loss

    def on_test_epoch_end(self):
        next_token_results = self.test_next_token_metric.compute()
        for metric_name, value in next_token_results.items():
            self.log(f"test/NEXT_TOKEN/{metric_name.upper()}", value, on_epoch=True)
        self.test_next_token_metric.reset()

    @classmethod
    def get_metadata_means(cls, metadata_df):
        if "values/sum" not in metadata_df or "values/n_occurrences" not in metadata_df:
            raise ValueError("Missing 'values/sum' and/or 'values/n_occurrences' columns in metadata_df")
        metadata_df = metadata_df.with_columns(
            (pl.col("values/sum") / pl.col("values/n_occurrences")).alias("values/mean")
        )
        return metadata_df

    @classmethod
    def get_code_to_time_map(cls, metadata_df) -> dict:
        """Convert the metadata DataFrame to a dictionary mapping code to time.

        Args:
            metadata_df: Polars DataFrame containing code metadata
                (includes 'code' and 'code/vocab_index' columns)

        Returns:
            dict: Mapping code to time in years

        Example:
        >>> metadata_df = pl.DataFrame({
        ...     "code": ["A", "B", "C", "TIME//DELTA//TOKEN//_Q_17"],
        ...     "code/vocab_index": [0, 1, 2, 3],
        ...     "values/sum": [None, None, None, 1],
        ...     "values/n_occurrences": [None, None, None, 1],
        ... })
        >>> EicForecastingModule.get_code_to_time_map(metadata_df)
        tensor([0., 0., 0., 1., 0.])
        """
        metadata_df = cls.get_metadata_means(metadata_df)
        # Assuming we know the vocab size
        num_vocab = metadata_df["code/vocab_index"].max()
        code_to_time_map = torch.zeros(num_vocab + 2)  # +2 since indices start at 1 and EOS token is added

        # Set values using the indices
        time_mask = pl.col("code").str.starts_with(TIME_DELTA_TOKEN)
        vocab_indices = metadata_df.filter(time_mask)["code/vocab_index"]
        time_values = metadata_df.filter(time_mask)["values/mean"]

        code_to_time_map[vocab_indices.to_torch().to(torch.long)] = time_values.to_torch().to(torch.float)
        return code_to_time_map

    @classmethod
    def get_code_to_numeric_value_map(cls, metadata_df, get_raw_values=True) -> dict:
        """Convert the metadata DataFrame to a dictionary mapping code to numeric value.

        Args:
            metadata_df: Polars DataFrame containing code metadata
                (includes 'code' and 'code/vocab_index' columns)

        Returns:
            dict: Mapping code to time in years

        Example:
        >>> metadata_df = pl.DataFrame({
        ...     "code": ["A", "A//_Q_1", "A//_Q_2", "A//_Q_3", "A//_Q_4", "B"],
        ...     "code/vocab_index": [0, 1, 2, 3, 4, 5],
        ...     'values/min': [0, 0, 0, 0, 0, None],
        ...     'values/max': [4, 4, 4, 4, 4, None],
        ...     'values/sum': [None, .5, 1.5, 2.5, 3.5, None],
        ...     'values/n_occurrences': [None, 1, 1, 1, 1, None],
        ...     "values/quantiles": [
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': None, 'values/quantile/0.5': None,
        ...          'values/quantile/0.75': None},
        ...     ],
        ... })
        >>> EicForecastingModule.get_code_to_numeric_value_map(metadata_df, get_raw_values=True).tolist()
        [nan, 0.5, 1.5, 2.5, 3.5, nan, nan]
        >>> EicForecastingModule.get_code_to_numeric_value_map(metadata_df, get_raw_values=False).tolist()
        [nan, 0.125, 0.375, 0.625, 0.875, nan, nan]
        """
        # First, verify the input DataFrame is sorted by vocab_index
        assert metadata_df["code/vocab_index"].is_sorted()

        # Get the maximum vocab index to determine tensor size
        max_vocab_idx = metadata_df["code/vocab_index"].max()

        # Create a tensor filled with NaN values
        result = torch.full((max_vocab_idx + 1,), float("nan"))
        # TODO(Oufattole) remove this and enforce that metadata_df must include the values/min
        ordered_quantiles = [field.name for field in metadata_df.schema["values/quantiles"].fields]
        percentiles = [0, *[float(q.split("/")[-1]) for q in ordered_quantiles], 1]
        if "values/min" not in metadata_df.columns or "values/max" not in metadata_df.columns:
            raise ValueError("Missing values/min and/or values/max values in metadata_df")
        metadata_df = cls.get_metadata_means(metadata_df)

        # Process each row in the DataFrame
        for row in metadata_df.iter_rows(named=True):
            vocab_idx = row["code/vocab_index"]
            code = row["code"]
            raw_quantiles = [row["values/quantiles"][each] for each in ordered_quantiles]
            min_value = row["values/min"]
            max_value = row["values/max"]
            raw_quantiles = [min_value, *raw_quantiles, max_value]
            mean_value = row["values/mean"]

            # Check if this is a quarterly code (contains "//_Q_")
            if code and "//_Q_" in code and not code.startswith("TIME//DELTA//TOKEN"):
                # Extract the number of quantiles the value is greater than, 0 for Q_1, 1 for Q_2, etc.
                rank = int(code.split("//_Q_")[1]) - 1
                # We estimate the numeric value is the average of the bordering quantiles it is between
                if get_raw_values:
                    result[vocab_idx] = mean_value
                    # result[vocab_idx] = sum([raw_quantiles[rank], raw_quantiles[rank + 1]]) / 2
                else:
                    result[vocab_idx] = sum([percentiles[rank], percentiles[rank + 1]]) / 2

            # For non-quarterly codes, leave as NaN
            # This handles both the base code (e.g., "A") and any other non-quarterly codes
        return torch.cat([result, torch.Tensor([np.nan])])  # postpend a zero in case EOS token is postpended

    @classmethod
    def to_trajectory_batch(
        cls,
        code,
        mask,
        metadata_df,
        prediction_time_offset_years: torch.Tensor,
        code_to_time_map: torch.Tensor = None,
        code_to_numeric_value_map: torch.Tensor = None,
    ) -> TrajectoryBatch:
        """Convert the model output to MEDS format.

        Args:
            code (torch.Tensor): Tensor of shape (batch_size, sequence_length) containing event codes
            mask (torch.Tensor): Tensor of shape (batch_size, sequence_length) indicates valid
                measurements/codes
            metadata_df: Polars DataFrame containing code metadata (includes 'code' column)
            prediction_time_offset_days: Tensor of shape (batch_size,) containing the time difference in days
                between each input sequence's end time and its target prediction time. Used to calculate
                absolute timestamps since the TrajectoryBatch stores times relative to the prediction time.

        Returns:
            pl.DataFrame: MEDS format DataFrame with columns:
                - time_index: Time in years starting from 0
                - code: The medical code
                - value: Always 1.0 (presence indicator)
                - sample_id: ID of the generated sample

        Time will start from 0, and is measured in years.

        Example:
        >>> from datetime import datetime
        >>> metadata_df = pl.DataFrame({
        ...     "code": ["A", "A//_Q_1", "A//_Q_2", "A//_Q_3", "A//_Q_4", "TIME//DELTA//TOKEN//_Q_17"],
        ...     "code/vocab_index": [0, 1, 2, 3, 4, 5],
        ...     'values/min': [0, 0, 0, 0, 0, None],
        ...     'values/max': [4, 4, 4, 4, 4, None],
        ...     'values/sum': [None, .5, 1.5, 2.5, 3.5, 1],
        ...     'values/n_occurrences': [None, 1, 1, 1, 1, 1],
        ...     "values/quantiles": [
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': 1, 'values/quantile/0.5': 2, 'values/quantile/0.75': 3},
        ...         {'values/quantile/0.25': None, 'values/quantile/0.5': None,
        ...          'values/quantile/0.75': None},
        ...     ],
        ... })
        >>> code = torch.tensor([[0, 2, 5, 5], [2, 3, 4, 5], [5, 5, 0, 1]])
        >>> mask = torch.tensor([[1, 1, 1, 1], [1, 1, 1, 0], [1, 1, 1, 0]])
        >>> prediction_time_offset_years = torch.tensor([0.0, 1.0, 2.0])
        >>> from pprint import pprint, pformat
        >>> subject_ids = [1,2,3]
        >>> prediction_times = [1,2,3]
        >>> EicForecastingModule.to_trajectory_batch(code, mask, metadata_df, prediction_time_offset_years
        ...     ).to_meds(subject_ids, prediction_times).columns
        ['subject_id', 'prediction_time', 'time', 'code', 'code/vocab_index', 'numeric_value']
        """
        if not code_to_time_map:
            code_to_time_map = cls.get_code_to_time_map(metadata_df)
        if not code_to_numeric_value_map:
            code_to_numeric_value_map = cls.get_code_to_numeric_value_map(metadata_df)
        # Initialize lists to store the DataFrame rows
        time = torch.cumsum(code_to_time_map[code], dim=1)
        numeric_value = code_to_numeric_value_map[code]
        numeric_value_mask = ~numeric_value.isnan()
        time += prediction_time_offset_years.unsqueeze(1)
        return TrajectoryBatch(time, code, mask, numeric_value, numeric_value_mask, metadata_df)

    def update_generation_state(
        self,
        tokens: torch.Tensor,
        cumulative_time: torch.Tensor,
        trajectory_labeler: SequenceLabeler | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor, bool]:
        """Updates trajectory_labeler state, and returns state information.

        Examples:
            >>> import tempfile
            >>> from clinical_zeroshot_labeler.labeler import WindowStatus
            >>> # Create test setup using helper function
            >>> _, metadata_df, _, _ = create_dummy_sequence_labeler()

            >>> # Write metadata to temporary file and create config
            >>> temp_file = tempfile.NamedTemporaryFile(suffix='.parquet')
            >>> metadata_df.write_parquet(temp_file.name)
            >>> cfg = create_model_config(temp_file.name)

            >>> model = EicForecastingModule(cfg)
            >>> model._init_time_and_value_quantiles()
            >>> B = 2  # batch_size
            >>> device = 'cpu'

            >>> # Setup basic test case
            >>> cumulative = torch.tensor([0.0, 0.0], device=device)
            >>> tokens = torch.randint(0, 5, (B,3), device=device)

            >>> # Test trajectory labeler progression
            >>> labeler = DummyTrajectoryLabeler(B)
            >>> time, status, is_finished, ended = model.update_generation_state(
            ...     tokens=tokens,
            ...     cumulative_time=cumulative,
            ...     trajectory_labeler=labeler,
            ... )
            >>> assert time.shape == (B,)
            >>> assert status.shape == (B,)
            >>> assert not is_finished
            >>> assert not ended.any()

            >>> # Test second step shows active status
            >>> time, status, is_finished, ended = model.update_generation_state(
            ...     tokens=tokens,
            ...     cumulative_time=time,
            ...     trajectory_labeler=labeler,
            ... )
            >>> assert (status == WindowStatus.ACTIVE.value).all()
            >>> assert not is_finished
            >>> assert not ended.any()

            >>> # Test third step shows satisfied status and finished
            >>> time, status, is_finished, ended = model.update_generation_state(
            ...     tokens=tokens,
            ...     cumulative_time=time,
            ...     trajectory_labeler=labeler,
            ... )
            >>> assert (status == WindowStatus.SATISFIED.value).all()
            >>> assert is_finished
            >>> assert ended.all()

            >>> # Test without trajectory labeler
            >>> time, status, is_finished, ended = model.update_generation_state(
            ...     tokens=tokens,
            ...     cumulative_time=time,
            ...     trajectory_labeler=None,
            ... )
            >>> assert time.shape == (B,)
            >>> assert status is None
            >>> assert not is_finished
            >>> assert not ended.any()
        """
        current_sample = tokens[:, -1].cpu()
        pred_time = self.time_quantile_map[current_sample.flatten()]
        cumulative_time = cumulative_time.cpu() + pred_time.squeeze(-1)
        current_value = self.value_quantile_map[current_sample.flatten()]
        if trajectory_labeler is not None:
            status = trajectory_labeler.process_step(current_sample, cumulative_time, current_value)
            is_finished = trajectory_labeler.is_finished()
            ended_sequences = torch.logical_or(
                status == WindowStatus.SATISFIED.value, status == WindowStatus.IMPOSSIBLE.value
            )
        else:
            status = None
            is_finished = False
            ended_sequences = torch.zeros((current_sample.shape[0]), dtype=torch.bool)
        return cumulative_time, status, is_finished, ended_sequences

    def _init_time_and_value_quantiles(self):
        if not hasattr(self, "time_quantile_map"):
            self.time_quantile_map = self.get_code_to_time_map(self.metadata_df)
        if not hasattr(self, "value_quantile_map"):
            self.value_quantile_map = self.get_code_to_numeric_value_map(self.metadata_df)

    @torch.inference_mode()
    @eval_decorator
    @TimeableMixin.TimeAs
    def generate_batch(
        self,
        input_batch,
        **kwargs,
    ):
        """Generate evaluation metrics for the model."""
        if self.cfg.max_tokens_budget is None and self.trajectory_labeler is None:
            raise ValueError(
                "At least one of model.backbone.max_tokens_budget or model.trajectory_labeler must be "
                "set in the configuration."
            )
        if self.cfg.backbone.cfg.token_emb:
            raise NotImplementedError(
                "Token embeddings not supported, use x-transformers library for token embeddings"
            )
        else:
            prompts, mask = input_batch[INPUT_ENCODER_TOKENS_KEY], input_batch[INPUT_ENCODER_MASK_KEY]

        self._init_time_and_value_quantiles()

        if "prediction_time" not in input_batch or "end_time" not in input_batch:
            raise ValueError(
                "Prediction time and end time must be provided for zero-shot labeling. "
                "Enable the flags do_include_prediction_time and do_include_end_time."
            )
        prediction_time_offset_years = (
            -get_time_days_delta(input_batch["prediction_time"], input_batch["end_time"], prompts.device)
            / 365.25
        )
        if (prediction_time_offset_years > 0).any():
            raise ValueError("time_offset_years must be less than or equal to 0")

        if self.cfg.generate_id is not None:
            trajectory_labeler = (
                self.trajectory_labeler(batch_size=prompts.shape[0], metadata_df=self.metadata_df)
                if self.trajectory_labeler is not None
                else None
            )
            out, out_lengths, metadata = self.generate(
                prompts=prompts,
                mask=mask,
                trajectory_labeler=trajectory_labeler,
                time_offset_years=prediction_time_offset_years,
                temperature=self.cfg.temperature,
                eos_tokens=self.cfg.eos_tokens,
                log_progress=self.cfg.get("log_progress", False),
                prune_terminated=self.cfg.prune_terminated,
                **kwargs,
            )
            out_mask = torch.arange(out.size(1))[None, :].cpu() < out_lengths[:, None].cpu()

            # Store generated data
            null_data = torch.zeros_like(out).cpu()
            # Convert codes to time deltas
            time_deltas = self.time_quantile_map.to(out.device)[out]
            generated_data = {
                "code": out.cpu(),
                "mask": out_mask,
                "numeric_value": null_data,
                "numeric_value_mask": null_data,
                "static_mask": null_data,
                "time_delta_years": time_deltas.cpu(),
                "subject_id": input_batch["subject_id"].cpu(),
                "prediction_time": input_batch["prediction_time"],
                "end_time": input_batch["end_time"],
            }
            trajectory_batch = self.to_trajectory_batch(
                generated_data["code"],
                generated_data["mask"],
                self.metadata_df,
                prediction_time_offset_years.cpu(),
            )
            if self.cfg.store_generated_trajectory:
                input_batch[GENERATE_PREFIX + str(self.cfg.generate_id)] = trajectory_batch.to_meds(
                    generated_data["prediction_time"], generated_data["subject_id"]
                )
            logger.info(f"Completed generation for sample {self.cfg.generate_id}")

            if metadata:
                labels, status = metadata["labels"], metadata["status"]
                input_batch[MODEL_PREFIX + "STATUS"] = status
                unknown = status != WindowStatus.SATISFIED.value
                # Handle unknown values by setting their probability to 0.5
                if unknown.any().item() > 0:
                    logger.warning(f"Found {unknown.sum().item()} unknown zero-shot predictions")
                    labels[unknown] = 0.5
                input_batch[MODEL_PRED_PROBA_KEY] = labels
                input_batch[MODEL_PRED_STATUS_KEY] = status
                input_batch["MODEL//OFFSET"] = prediction_time_offset_years
                logger.info(f"Completed zero-shot labeling for sample {self.cfg.generate_id}")
            if trajectory_labeler is not None and self.cfg.return_labeler:
                input_batch["labeler"] = trajectory_labeler
        return input_batch
